from typing import Callable, Dict, Optional, Sequence, Union
from hub.util.dataset import try_flushing
from hub.util.dataset import map_tensor_keys
from .common import (
    PytorchTransformFunction,
    convert_fn as default_convert_fn,
    collate_fn as default_collate_fn,
)
from hub.util.exceptions import EmptyTensorError


def create_dataloader_nesteddataloader(
    dataset,
    tensors,
    tobytes,
    use_local_cache,
    transform,
    num_workers,
    buffer_size,
    batch_size,
    collate_fn,
    pin_memory,
    drop_last,
    return_index,
    pad_tensors,
):
    import torch
    import torch.utils.data
    from hub.integrations.pytorch.dataset import SubIterableDataset

    return torch.utils.data.DataLoader(
        # this data set is more efficient also shuffles
        # using threads race conditions as source of entropy
        SubIterableDataset(
            dataset,
            tensors=tensors,
            tobytes=tobytes,
            use_local_cache=use_local_cache,
            transform=transform,
            batch_size=batch_size,
            num_workers=num_workers,
            buffer_size=buffer_size,
            return_index=return_index,
            pad_tensors=pad_tensors,
        ),
        batch_size=batch_size,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
        drop_last=drop_last,
    )


def create_dataloader_shufflingdataloader(
    dataset,
    tensors,
    tobytes,
    use_local_cache,
    transform,
    num_workers,
    buffer_size,
    batch_size,
    collate_fn,
    pin_memory,
    drop_last,
):
    import torch
    import torch.utils.data
    from hub.integrations.pytorch.dataset import ShufflingIterableDataset

    return torch.utils.data.DataLoader(
        # this data set is more efficient also shuffles
        # using threads race conditions as source of entropy
        ShufflingIterableDataset(
            dataset,
            tensors=tensors,
            tobytes=tobytes,
            use_local_cache=use_local_cache,
            transform=transform,
            num_workers=num_workers,
            buffer_size=buffer_size,
        ),
        batch_size=batch_size,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
        drop_last=drop_last,
    )


create_dataloader = create_dataloader_nesteddataloader


def dataset_to_pytorch(
    dataset,
    num_workers: int,
    batch_size: int,
    drop_last: bool,
    collate_fn: Optional[Callable],
    pin_memory: bool,
    shuffle: bool,
    buffer_size: int,
    use_local_cache: bool,
    transform: Optional[Union[Dict, Callable]] = None,
    tensors: Optional[Sequence[str]] = None,
    tobytes: Union[bool, Sequence[str]] = False,
    return_index: bool = True,
    pad_tensors: bool = True,
):

    import torch
    from hub.integrations.pytorch.dataset import TorchDataset

    try_flushing(dataset)

    torch.multiprocessing.set_sharing_strategy("file_system")

    if collate_fn is None:
        collate_fn = default_convert_fn if batch_size is None else default_collate_fn

    if tensors is not None and "index" in tensors:
        raise ValueError("index is not a tensor, to get index, pass return_index=True")

    tensors = map_tensor_keys(dataset, tensors)
    if isinstance(transform, dict):
        tensors = [k for k in transform.keys() if k != "index"]
        transform = PytorchTransformFunction(transform_dict=transform)
    else:
        transform = PytorchTransformFunction(composite_transform=transform)

    # check whether we have an empty tensor inside of tensors
    for tensor_name in tensors:
        tensor = dataset._get_tensor_from_root(tensor_name)
        if len(tensor) == 0:
            raise EmptyTensorError(
                f" the dataset has an empty tensor {tensor_name}, pytorch dataloader can't be created."
                f" Please either populate the tensor or pass tensors argument to .pytorch that excludes this"
                f" tensor."
            )

    if shuffle and num_workers > 0:
        return create_dataloader(
            dataset,
            tensors,
            tobytes,
            use_local_cache,
            transform,
            num_workers,
            buffer_size,
            batch_size,
            collate_fn,
            pin_memory,
            drop_last,
            return_index,
            pad_tensors,
        )
    else:
        return torch.utils.data.DataLoader(
            TorchDataset(
                dataset,
                tensors=tensors,
                tobytes=tobytes,
                use_local_cache=use_local_cache,
                transform=transform,
                num_workers=num_workers,
                shuffle=shuffle,
                buffer_size=buffer_size,
                return_index=return_index,
                pad_tensors=pad_tensors,
            ),
            batch_size=batch_size,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
            num_workers=num_workers,
            drop_last=drop_last,
        )



from typing import List, Union, Sequence, Any
from functools import reduce
import numpy as np
from hub.core.linked_sample import LinkedSample
from hub.util.exceptions import TensorDtypeMismatchError
from hub.core.sample import Sample  # type: ignore
import hub


def _get_bigger_dtype(d1, d2):
    if np.can_cast(d1, d2):
        if np.can_cast(d2, d1):
            return d1
        else:
            return d2
    else:
        if np.can_cast(d2, d1):
            return d2
        else:
            return np.object


def get_dtype(val: Union[np.ndarray, Sequence, Sample]) -> np.dtype:
    """Get the dtype of a non-uniform mixed dtype sequence of samples."""

    if hasattr(val, "dtype"):
        return np.dtype(val.dtype)  # type: ignore
    elif isinstance(val, int):
        return np.array(0).dtype
    elif isinstance(val, float):
        return np.array(0.0).dtype
    elif isinstance(val, (str, LinkedSample)):
        return np.array("").dtype
    elif isinstance(val, bool):
        return np.dtype(bool)
    elif isinstance(val, Sequence):
        return reduce(_get_bigger_dtype, map(get_dtype, val))
    else:
        raise TypeError(f"Cannot infer numpy dtype for {val}")


def get_htype(val: Union[np.ndarray, Sequence, Sample]) -> str:
    """Get the htype of a non-uniform mixed dtype sequence of samples."""
    if isinstance(val, hub.core.tensor.Tensor):
        return val.meta.htype
    if hasattr(val, "shape"):  # covers numpy arrays, numpy scalars and hub samples.
        return "generic"
    if isinstance(val, List) and len(val) > 0 and isinstance(val[0], LinkedSample):
        return "generic"
    types = set((map(type, val)))  # type: ignore
    if dict in types:
        return "json"
    if types == set((str,)):
        return "text"
    if object in [  # type: ignore
        np.array(x).dtype if not isinstance(x, np.ndarray) else x.dtype for x in val  # type: ignore
    ]:
        return "json"
    return "generic"


def get_empty_text_like_sample(htype: str):
    """Get an empty sample of the given htype.

    Args:
        htype: htype of the sample.

    Returns:
        Empty sample.

    Raises:
        ValueError: if htype is not one of 'text', 'json', and 'list'.
    """
    if htype == "text":
        return ""
    elif htype == "json":
        return {}
    elif htype == "list":
        return []
    else:
        raise ValueError(
            f"This method should only be used for htypes 'text', 'json' and 'list'. Got {htype}."
        )


def intelligent_cast(
    sample: Any, dtype: Union[np.dtype, str], htype: str
) -> np.ndarray:
    # TODO: docstring (note: sample can be a scalar)/statictyping
    # TODO: implement better casting here
    if isinstance(sample, Sample):
        sample = sample.array

    if hasattr(sample, "dtype") and sample.dtype == dtype:
        return sample

    err_dtype = get_incompatible_dtype(sample, dtype)
    if err_dtype:
        raise TensorDtypeMismatchError(
            dtype,
            err_dtype,
            htype,
        )

    if hasattr(sample, "astype"):  # covers both ndarrays and scalars
        return sample.astype(dtype)

    return np.array(sample, dtype=dtype)


def get_incompatible_dtype(
    samples: Union[np.ndarray, Sequence], dtype: Union[str, np.dtype]
):
    """Check if items in a non-uniform mixed dtype sequence of samples can be safely cast to the given dtype.

    Args:
        samples: Sequence of samples
        dtype: dtype to which samples have to be cast

    Returns:
        None if all samples are compatible. If not, the dtype of the offending item is returned.

    Raises:
        TypeError: if samples is of unexepcted type.
    """
    if isinstance(samples, np.ndarray):
        if samples.size == 0:
            return None
        elif samples.size == 1:
            samples = samples.reshape(1).tolist()[0]

    if isinstance(samples, (int, float, bool)) or hasattr(samples, "dtype"):
        return (
            None
            if np.can_cast(samples, dtype)
            else getattr(samples, "dtype", np.array(samples).dtype)
        )
    elif isinstance(samples, Sequence):
        for dt in map(lambda x: get_incompatible_dtype(x, dtype), samples):
            if dt:
                return dt
        return None
    else:
        raise TypeError(
            f"Unexpected object {samples}. Expected np.ndarray, int, float, bool, str or Sequence."
        )



from hub.constants import MB, KB, PARTIAL_NUM_SAMPLES
from hub.core.chunk.chunk_compressed_chunk import ChunkCompressedChunk
import numpy as np
import pytest

import hub
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.sample import Sample  # type: ignore
from hub.core.tiling.deserialize import np_list_to_sample
from hub.core.tiling.sample_tiles import SampleTiles

compressions_paremetrized = pytest.mark.parametrize("compression", ["lz4", "png"])

common_args = {
    "min_chunk_size": 1 * MB,
    "max_chunk_size": 2 * MB,
    "tiling_threshold": 1 * MB,
}


def create_tensor_meta():
    tensor_meta = TensorMeta()
    tensor_meta.dtype = "uint8"
    tensor_meta.max_shape = None
    tensor_meta.min_shape = None
    tensor_meta.htype = "generic"
    tensor_meta.length = 0
    return tensor_meta


@compressions_paremetrized
def test_read_write_sequence(compression):
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    common_args["compression"] = compression
    dtype = tensor_meta.dtype
    data_in = [
        np.random.randint(0, 255, size=(250, 125)).astype(dtype) for _ in range(10)
    ]
    data_in2 = data_in.copy()
    while data_in:
        chunk = ChunkCompressedChunk(**common_args)
        num_samples = int(chunk.extend_if_has_space(data_in))
        chunk._decompressed_samples = None
        data_out = [chunk.read_sample(i) for i in range(num_samples)]
        np.testing.assert_array_equal(data_out, data_in2[:num_samples])
        data_in = data_in[num_samples:]
        data_in2 = data_in2[num_samples:]


@compressions_paremetrized
@pytest.mark.parametrize("random", [True, False])
def test_read_write_sequence_big(cat_path, compression, random):
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    common_args["compression"] = compression
    dtype = tensor_meta.dtype
    data_in = []
    for i in range(50):
        if i % 10 == 0:
            data_in.append(
                np.random.randint(0, 255, size=(1501, 750, 3)).astype(dtype) * random
            )
        elif i % 3 == 0:
            data_in.append(
                hub.read(cat_path) if random else np.zeros((225, 225, 3), dtype=dtype)
            )
        else:
            data_in.append(
                np.random.randint(0, 255, size=(250, 125, 3)).astype(dtype) * random
            )
    data_in2 = data_in.copy()
    tiles = []
    original_length = len(data_in)
    tiled = False
    while data_in:
        chunk = ChunkCompressedChunk(**common_args)
        chunk._compression_ratio = 10  # start with a bad compression ratio to hit exponential back off code path
        num_samples = chunk.extend_if_has_space(data_in)
        if num_samples == PARTIAL_NUM_SAMPLES:
            tiled = True
            tiles.append(chunk.read_sample(0, is_tile=True))
            sample = data_in[0]
            assert isinstance(sample, SampleTiles)
            if sample.is_last_write:
                current_length = len(data_in)
                index = original_length - current_length
                full_data_out = np_list_to_sample(
                    tiles,
                    sample.sample_shape,
                    sample.tile_shape,
                    sample.layout_shape,
                    dtype,
                )
                np.testing.assert_array_equal(full_data_out, data_in2[index])
                data_in = data_in[1:]
                tiles = []

        elif num_samples > 0:
            data_out = [chunk.read_sample(i) for i in range(num_samples)]
            for i, item in enumerate(data_out):
                if isinstance(item, Sample):
                    item = item.array
                np.testing.assert_array_equal(item, data_in[i])
            data_in = data_in[num_samples:]
    assert tiled


@compressions_paremetrized
def test_update(compression):
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    common_args["compression"] = compression
    dtype = tensor_meta.dtype
    arr = np.random.randint(0, 255, size=(7, 75, 50, 3)).astype(dtype)
    data_in = list(arr)
    chunk = ChunkCompressedChunk(**common_args)
    chunk.extend_if_has_space(data_in)

    data_out = np.array([chunk.read_sample(i) for i in range(7)])
    np.testing.assert_array_equal(data_out, data_in)

    data_3 = np.random.randint(0, 255, size=(175, 350, 3)).astype(dtype)
    data_5 = np.random.randint(0, 255, size=(500, 750, 3)).astype(dtype)

    chunk.update_sample(3, data_3)
    chunk.update_sample(5, data_5)
    for i in range(7):
        if i == 3:
            np.testing.assert_array_equal(chunk.read_sample(i), data_3)
        elif i == 5:
            np.testing.assert_array_equal(chunk.read_sample(i), data_5)
        else:
            np.testing.assert_array_equal(chunk.read_sample(i), data_in[i])



import numpy as np
from typing import List, Union
from hub.core.serialize import check_sample_shape, bytes_to_text
from hub.core.tiling.sample_tiles import SampleTiles
from hub.util.casting import intelligent_cast
from hub.util.exceptions import EmptyTensorError
from .base_chunk import BaseChunk, InputSample


class UncompressedChunk(BaseChunk):
    def extend_if_has_space(  # type: ignore
        self,
        incoming_samples: Union[List[InputSample], np.ndarray],
        update_tensor_meta: bool = True,
    ) -> float:
        self.prepare_for_write()
        if isinstance(incoming_samples, np.ndarray):
            if incoming_samples.dtype == object:
                incoming_samples = list(incoming_samples)
            else:
                return self._extend_if_has_space_numpy(
                    incoming_samples, update_tensor_meta
                )
        return self._extend_if_has_space_list(incoming_samples, update_tensor_meta)

    def _extend_if_has_space_numpy(
        self,
        incoming_samples: np.ndarray,
        update_tensor_meta: bool = True,
    ) -> float:
        num_samples: int = 0
        buffer_size = 0

        for sample in incoming_samples:
            shape = self.normalize_shape(sample.shape)
            self.num_dims = self.num_dims or len(shape)
            sample_nbytes = sample.nbytes
            check_sample_shape(shape, self.num_dims)
            # no need to check sample size, this code is only reached when size smaller than chunk
            if not self.can_fit_sample(sample_nbytes, buffer_size):
                break
            buffer_size += sample_nbytes
            num_samples += 1

        samples = incoming_samples[:num_samples]
        samples = intelligent_cast(samples, self.dtype, self.htype)

        assert isinstance(self.data_bytes, bytearray)
        self.data_bytes += samples.tobytes()

        if num_samples > 0:
            shape = self.normalize_shape(samples[0].shape)
            sample_nbytes = samples[0].nbytes
            for _ in range(num_samples):
                self.register_in_meta_and_headers(
                    sample_nbytes, shape, update_tensor_meta=update_tensor_meta
                )

        return float(num_samples)

    def _extend_if_has_space_list(
        self,
        incoming_samples: List[InputSample],
        update_tensor_meta: bool = True,
    ) -> float:
        num_samples: float = 0

        for i, incoming_sample in enumerate(incoming_samples):
            serialized_sample, shape = self.serialize_sample(incoming_sample)
            if shape is not None:
                self.num_dims = self.num_dims or len(shape)
                check_sample_shape(shape, self.num_dims)

            # NOTE re-chunking logic should not reach to this point, for Tiled ones we do not have this logic
            if isinstance(serialized_sample, SampleTiles):
                incoming_samples[i] = serialized_sample  # type: ignore
                if self.is_empty:
                    self.write_tile(serialized_sample)
                    num_samples += 0.5
                break
            else:
                sample_nbytes = len(serialized_sample)
                if self.is_empty or self.can_fit_sample(sample_nbytes):
                    self.data_bytes += serialized_sample  # type: ignore

                    self.register_in_meta_and_headers(
                        sample_nbytes,
                        shape,
                        update_tensor_meta=update_tensor_meta,
                    )
                    num_samples += 1
                else:
                    break

        return num_samples

    def read_sample(
        self,
        local_index: int,
        cast: bool = True,
        copy: bool = False,
        decompress: bool = True,
        is_tile: bool = False,
    ):
        if self.is_empty_tensor:
            raise EmptyTensorError(
                "This tensor has only been populated with empty samples. "
                "Need to add at least one non-empty sample before retrieving data."
            )
        partial_sample_tile = self._get_partial_sample_tile()
        if partial_sample_tile is not None:
            return partial_sample_tile
        buffer = self.memoryview_data
        if not is_tile and self.is_fixed_shape:
            shape = tuple(self.tensor_meta.min_shape)
            sb, eb = self.get_byte_positions(local_index)
            buffer = buffer[sb:eb]
        else:
            shape = self.shapes_encoder[local_index]
            if not self.byte_positions_encoder.is_empty():
                sb, eb = self.byte_positions_encoder[local_index]
                buffer = buffer[sb:eb]

        if not decompress:
            if copy:
                buffer = bytes(buffer)
            return buffer
        if self.is_text_like:
            buffer = bytes(buffer)
            return bytes_to_text(buffer, self.htype)
        ret = np.frombuffer(buffer, dtype=self.dtype).reshape(shape)
        if copy and not ret.flags["WRITEABLE"]:
            ret = ret.copy()
        return ret

    def update_sample(self, local_index: int, sample: InputSample):
        self.prepare_for_write()
        serialized_sample, shape = self.serialize_sample(sample, break_into_tiles=False)
        self.check_shape_for_update(shape)
        new_nb = (
            None if self.byte_positions_encoder.is_empty() else len(serialized_sample)
        )

        old_data = self.data_bytes
        self.data_bytes = self.create_updated_data(
            local_index, old_data, serialized_sample
        )
        self.update_in_meta_and_headers(local_index, new_nb, shape)



from hub.core.linked_sample import LinkedSample
from typing import Optional, Dict


def link(
    path: str,
    creds_key: Optional[str] = None,
) -> LinkedSample:
    """Utility that stores a link to raw data. Used to add data to a Hub Dataset without copying it. See :ref:`Link htype`.

    Supported file types::

        Image: "bmp", "dib", "gif", "ico", "jpeg", "jpeg2000", "pcx", "png", "ppm", "sgi", "tga", "tiff", "webp", "wmf", "xbm"
        Audio: "flac", "mp3", "wav"
        Video: "mp4", "mkv", "avi"
        Dicom: "dcm"

    Args:
        path (str): Path to a supported file.
        creds_key (optional, str): The credential key to use to read data for this sample. The actual credentials are fetched from the dataset.

    Returns:
        LinkedSample: LinkedSample object that stores path and creds.
    """
    return LinkedSample(path, creds_key)



import numpy as np
from hub.util import split


def test_split(memory_ds):
    ds = memory_ds
    ds.create_tensor("ints", dtype="int64")
    ds.ints.extend(np.arange(13, dtype="int64").reshape((13, 1)))

    train, test, val = split(ds, [0.7, 0.2, 0.1])

    expected_train = [[i] for i in range(9)]
    expected_test = [[i + 9] for i in range(2)]
    expected_val = [[i + 11] for i in range(2)]

    assert train.ints.numpy().tolist() == expected_train
    assert test.ints.numpy().tolist() == expected_test
    assert val.ints.numpy().tolist() == expected_val

    assert len(train) == 9
    assert len(test) == 2
    assert len(val) == 2
    assert sum(map(len, (train, test, val))) == 13



import numpy as np
import hub
from hub.api.tests.test_api_tiling import compressions_paremetrized
import pytest


def pop_helper_basic(ds, pop_count):
    for i in range(len(ds.xyz)):
        ofs = 1 if i < 5 else 1 + pop_count
        target = i + ofs
        assert ds.xyz[i].shape == ds.xyz[i].numpy().shape == (target, target)
        np.testing.assert_array_equal(
            ds.xyz[i].numpy(), target * np.ones((target, target))
        )


def pop_helper_link(ds):
    assert len(ds.xyz) == 9
    for i in range(9):
        target = (513, 464, 4) if i % 2 == 0 else (900, 900, 3)
        assert ds.xyz[i].numpy().shape == ds.xyz[i].shape == target


def test_multiple(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("xyz")
        for i in range(1, 11):
            ds.xyz.append(i * np.ones((i, i)))
        for pop_count in range(1, 6):
            ds.xyz.pop(5)
            assert len(ds.xyz) == 10 - pop_count
            pop_helper_basic(ds, pop_count)

    ds = local_ds_generator()
    pop_helper_basic(ds, 5)

    with ds:
        ds.xyz.append(20 * np.ones((20, 20)))

    assert len(ds) == 6
    np.testing.assert_array_equal(ds.xyz[5].numpy(), 20 * np.ones((20, 20)))

    for i in range(6):
        ds.xyz.pop(0)
    assert len(ds) == 0
    assert ds.xyz.meta.max_shape == []
    assert ds.xyz.meta.min_shape == []

    ds = local_ds_generator()
    assert len(ds) == 0
    assert ds.xyz.meta.max_shape == []
    assert ds.xyz.meta.min_shape == []

    with ds:
        ds.xyz.append(30 * np.ones((30, 30)))

    assert len(ds) == 1
    np.testing.assert_array_equal(ds.xyz[0].numpy(), 30 * np.ones((30, 30)))


def test_link_pop(local_ds_generator, cat_path, flower_path):
    with local_ds_generator() as ds:
        ds.create_tensor("xyz", htype="link[image]")
        for i in range(10):
            url = cat_path if i % 2 == 0 else flower_path
            ds.xyz.append(hub.link(url))
        assert ds.xyz[0].numpy().shape == ds.xyz[0].shape == (900, 900, 3)
        ds.xyz.pop(0)
        assert len(ds.xyz) == 9
        pop_helper_link(ds)

    ds = local_ds_generator()
    assert len(ds.xyz) == 9
    pop_helper_link(ds)

    ds.xyz.append(hub.link(cat_path))
    assert ds.xyz[9].numpy().shape == ds.xyz[9].shape == (900, 900, 3)


def test_tiling_pop(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("xyz")
        ds.xyz.append(np.ones((1000, 1000, 3)))
        ds.xyz.append(2 * np.ones((2000, 2000, 3)))
        ds.xyz.append(3 * np.ones((3000, 3000, 3)))
        ds.xyz.pop(1)
        assert len(ds.xyz) == 2

        assert ds.xyz[0].numpy().shape == ds.xyz[0].shape == (1000, 1000, 3)
        np.testing.assert_array_equal(ds.xyz[0].numpy(), np.ones((1000, 1000, 3)))
        assert ds.xyz[1].numpy().shape == ds.xyz[1].shape == (3000, 3000, 3)
        np.testing.assert_array_equal(ds.xyz[1].numpy(), 3 * np.ones((3000, 3000, 3)))


@compressions_paremetrized
def test_compressions_pop(local_ds_generator, compression):
    ds = local_ds_generator()
    rint = np.random.randint
    ls = [
        rint(0, 255, (rint(50, 100), rint(50, 100), 3), dtype=np.uint8)
        for _ in range(6)
    ]
    with ds:
        ds.create_tensor("xyz", **compression)
        for i in range(5):
            ds.xyz.append(ls[i])

        ds.xyz.pop(2)
        assert len(ds.xyz) == 4
        np.testing.assert_array_equal(ds.xyz[0].numpy(), ls[0])
        np.testing.assert_array_equal(ds.xyz[1].numpy(), ls[1])
        np.testing.assert_array_equal(ds.xyz[2].numpy(), ls[3])
        np.testing.assert_array_equal(ds.xyz[3].numpy(), ls[4])

        ds.xyz.pop(0)
        assert len(ds.xyz) == 3
        np.testing.assert_array_equal(ds.xyz[0].numpy(), ls[1])
        np.testing.assert_array_equal(ds.xyz[1].numpy(), ls[3])
        np.testing.assert_array_equal(ds.xyz[2].numpy(), ls[4])

    ds = local_ds_generator()
    assert len(ds.xyz) == 3
    np.testing.assert_array_equal(ds.xyz[0].numpy(), ls[1])
    np.testing.assert_array_equal(ds.xyz[1].numpy(), ls[3])
    np.testing.assert_array_equal(ds.xyz[2].numpy(), ls[4])

    with ds:
        ds.xyz.append(ls[5])
        assert len(ds.xyz) == 4
        np.testing.assert_array_equal(ds.xyz[0].numpy(), ls[1])
        np.testing.assert_array_equal(ds.xyz[1].numpy(), ls[3])
        np.testing.assert_array_equal(ds.xyz[2].numpy(), ls[4])
        np.testing.assert_array_equal(ds.xyz[3].numpy(), ls[5])


def test_sequence_pop(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("xyz", htype="sequence")
        ds.xyz.append([[1, 2, 3], [4, 5, 6, 7]])
        ds.xyz.append([[8, 9], [11, 12, 13, 14]])
        ds.xyz.pop(0)
        assert len(ds.xyz) == 1
        val = ds.xyz[0].numpy(aslist=True)
        assert len(val) == 2
        assert (val[0] == [8, 9]).all()
        assert (val[1] == [11, 12, 13, 14]).all()

    ds = local_ds_generator()
    assert len(ds.xyz) == 1
    val = ds.xyz[0].numpy(aslist=True)
    assert len(val) == 2
    assert (val[0] == [8, 9]).all()
    assert (val[1] == [11, 12, 13, 14]).all()


def test_diff_pop(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("abc")
        a = ds.commit("first commit")
        for i in range(5):
            ds.abc.append(i)

        b = ds.commit("added 5 samples")

        ds.abc[2] = -2
        ds.abc[3] = -3
        ds.abc.pop(3)
        ds.abc.append(5)
        ds.abc.append(6)

        c = ds.commit("second commit")
        ds.abc.pop(2)

        diff1, diff2 = ds.diff(b, as_dict=True)["tensor"]
        assert diff1 == {
            "abc": {
                "created": False,
                "cleared": False,
                "info_updated": False,
                "data_transformed_in_place": False,
                "data_added": [3, 5],
                "data_updated": set(),
                "data_deleted": {2, 3},
            }
        }
        assert diff2 == {}


def test_ds_pop(local_ds):
    with local_ds as ds:
        ds.create_tensor("images")
        ds.create_tensor("labels")

        with pytest.raises(IndexError):
            ds.pop()

        for i in range(100):
            ds.images.append(i * np.ones((i + 1, i + 1, 3)))
            if i < 50:
                ds.labels.append(i)

        ds.pop(80)  # doesn't pop from tensors shorter than length 80
        assert len(ds.images) == 99
        assert len(ds.labels) == 50

        ds.pop(20)
        assert len(ds.images) == 98
        assert len(ds.labels) == 49

        ds.pop()  # only pops from the longest tensor
        assert len(ds.images) == 97
        assert len(ds.labels) == 49

        with pytest.raises(IndexError):
            ds.pop(-5)



from hub.util.token import expires_in_to_expires_at, is_expired_token
from time import sleep


def test_expiry():
    creds = {"Authorization": "Bearer 12345", "expires_in": 2}
    expires_in_to_expires_at(creds)
    sleep(2)
    assert is_expired_token(creds)



from typing import Dict, Optional, Union
import uuid
from flask import Flask, request, Response  # type: ignore
from hub.core.storage.provider import StorageProvider
from hub.util.threading import terminate_thread
from hub.client.config import (
    USE_DEV_ENVIRONMENT,
    USE_STAGING_ENVIRONMENT,
    USE_LOCAL_HOST,
)
import logging
import re
import socketserver
import threading

from IPython.display import IFrame, display  # type: ignore

_SERVER_THREAD: Optional[threading.Thread] = None
_APP = Flask("dataset_visualizer")

log = logging.getLogger("werkzeug")
log.setLevel(logging.ERROR)


def _run_app(port: int):
    try:
        _APP.run(host="0.0.0.0", port=port, threaded=True)
    except Exception:
        pass


@_APP.after_request
def after_request(response):
    response.headers.add("Accept-Ranges", "bytes")
    response.headers.add("Access-Control-Allow-Origin", "*")
    response.headers.add("Access-Control-Allow-Headers", "*")
    return response


class _Visualizer:
    """
    Visualizer class to manage visualization of the datasets.
    """

    _port: Optional[int] = None
    _storages: Dict = {}

    def __init__(self):
        self.start_server()
        self._storages = {}

    def add(self, storage: StorageProvider) -> str:
        id = str(uuid.uuid4())
        self._storages[id] = storage
        return id

    def get(self, id: str) -> StorageProvider:
        return self._storages[id]

    @property
    def port(self):
        return self._port

    def get_free_port(self):
        with socketserver.TCPServer(("localhost", 0), None) as s:
            return s.server_address[1]

    def is_server_running(self) -> bool:
        return (_SERVER_THREAD is not None) and _SERVER_THREAD.is_alive()

    def start_server(self):
        global _SERVER_THREAD
        if self.is_server_running():
            return f"http://localhost:{self.port}/"
        self._port = self.get_free_port()

        def run_app():
            _run_app(port=self.port)

        _SERVER_THREAD = threading.Thread(target=run_app, daemon=True)
        _SERVER_THREAD.start()
        return f"http://localhost:{self.port}/"

    def stop_server(self):
        global _SERVER_THREAD
        if not self.is_server_running():
            return
        terminate_thread(_SERVER_THREAD)
        _SERVER_THREAD = None

    def __del__(self):
        self.stop_server()


visualizer = _Visualizer()


def _get_visualizer_backend_url():
    if USE_LOCAL_HOST:
        return "http://localhost:3000"
    elif USE_DEV_ENVIRONMENT:
        return "https://app-dev.activeloop.dev"
    elif USE_STAGING_ENVIRONMENT:
        return "https://app-staging.activeloop.dev"
    else:
        return "https://app.activeloop.ai"


def visualize(
    source: Union[StorageProvider, str],
    token: Union[str, None] = None,
    width: Union[int, str, None] = None,
    height: Union[int, str, None] = None,
):
    """
    Visualizes the given dataset in the Jupyter notebook.

    Args:
        source: Union[StorageProvider, str] The storage or the path of the dataset.
        token: Union[str, None] Optional token to use in the backend call.
        width: Union[int, str, None] Optional width of the visualizer canvas.
        height: Union[int, str, None] Optional height of the visualizer canvas.
    """
    if isinstance(source, StorageProvider):
        id = visualizer.add(source)
        params = f"url=http://localhost:{visualizer.port}/{id}/"
    elif token is None:
        params = f"url={source}"
    else:
        params = f"url={source}&token={token}"
    iframe = IFrame(
        f"{_get_visualizer_backend_url()}/visualizer/hub?{params}",
        width=width or "90%",
        height=height or 800,
    )
    display(iframe)


@_APP.route("/<path:path>")
def access_data(path):
    try:
        paths = path.split("/", 1)
        range_header = request.headers.get("Range", None)
        start, end = 0, None
        storage: StorageProvider = visualizer.get(paths[0])
        if request.method == "HEAD":
            if paths[1] in storage.keys():
                return Response("OK", 200)
            else:
                return Response("", 404)
        if range_header:
            match = re.search(r"(\d+)-(\d*)", range_header)
            groups = match.groups()

            if groups[0]:
                start = int(groups[0])
            if groups[1]:
                end = int(groups[1]) + 1

        c = storage.get_bytes(paths[1], start, end)
        if isinstance(c, memoryview):
            c = c.tobytes()
        resp = Response(
            c,
            206,
            content_type="application/octet-stream",
        )
        resp.headers.add("Connection", "keep-alive")
        resp.headers.add("Accept-Ranges", "bytes")
        resp.headers.add("Content-Range", "bytes {0}-{1}".format(start, end))
        return resp

    except Exception as e:
        return Response(
            "Not Found",
            404,
            content_type="application/octet-stream",
        )



import numpy as np
import pytest

from hub.util.exceptions import (
    MergeConflictError,
    MergeMismatchError,
    MergeNotSupportedError,
)


def test_merge(local_ds):
    with local_ds as ds:
        ds.create_tensor("image")
        ds.image.append(1)
        a = ds.commit()
        ds.image[0] = 2
        b = ds.commit()
        assert ds.image[0].numpy() == 2
        ds.checkout(a)
        ds.checkout("alt", create=True)
        ds.image[0] = 3
        assert ds.image[0].numpy() == 3
        f = ds.commit()

        ds.checkout("main")
        assert ds.image[0].numpy() == 2

        ds.merge(f, conflict_resolution="theirs")
        assert ds.image[0].numpy() == 3

        ds.image[0] = 4
        assert ds.image[0].numpy() == 4
        d = ds.commit()
        ds.checkout("alt")
        assert ds.image[0].numpy() == 3

        ds.image[0] = 0
        assert ds.image[0].numpy() == 0
        g = ds.commit()

        ds.merge("main", conflict_resolution="theirs")
        assert ds.image[0].numpy() == 4

        ds.image[0] = 5
        assert ds.image[0].numpy() == 5
        ds.image.append(10)
        i = ds.commit()

        ds.image[0] = 6
        assert ds.image[0].numpy() == 6

        ds.checkout("main")
        assert ds.image[0].numpy() == 4

        ds.merge("alt", conflict_resolution="theirs")
        assert ds.image[0].numpy() == 6
        assert ds.image[1].numpy() == 10


def test_complex_merge(local_ds):
    with local_ds as ds:
        ds.create_tensor("image")
        ds.create_tensor("label")
        for i in range(10):
            ds.image.append(i * np.ones((200, 200, 3)))
            ds.label.append(i)
        a = ds.commit("added 10 images and labels")
        ds.checkout("other", create=True)
        for i in range(10, 15):
            ds.image.append(i * np.ones((200, 200, 3)))
        for i in range(3, 7):
            ds.label[i] = 2 * i

        b = ds.commit("added 5 more images and changed 4 labels")
        assert len(ds.image) == 15
        assert len(ds.label) == 10
        commit_id = ds.commit_id
        ds.merge("main")
        assert len(ds.image) == 15
        assert len(ds.label) == 10
        assert ds.commit_id == commit_id

        ds.checkout("main")
        ds.merge("other")
        assert len(ds.image) == 15
        assert len(ds.label) == 10
        for i in range(10):
            target = 1 if i not in range(3, 7) else 2
            assert ds.label[i].numpy() == target * i
        for i in range(15):
            np.testing.assert_array_equal(
                ds.image[i].numpy(), i * np.ones((200, 200, 3))
            )

        ds.checkout("other")
        ds.merge("main")

        assert len(ds.image) == 15
        assert len(ds.label) == 10
        for i in range(10):
            target = 1 if i not in range(3, 7) else 2
            assert ds.label[i].numpy() == target * i
        for i in range(15):
            np.testing.assert_array_equal(
                ds.image[i].numpy(), i * np.ones((200, 200, 3))
            )


def test_merge_not_supported(local_ds):
    with local_ds as ds:
        ds.create_tensor("image", create_id_tensor=False)
        ds.create_tensor("label")
        for i in range(10):
            ds.image.append(i * np.ones((200, 200, 3)))
            ds.label.append(i)
        a = ds.commit("added 10 images and labels")
        ds.checkout("other", create=True)
        for i in range(10, 15):
            ds.image.append(i * np.ones((200, 200, 3)))
        for i in range(3, 7):
            ds.label[i] = 2 * i

        b = ds.commit("added 5 more images and changed 4 labels")
        assert len(ds.image) == 15
        assert len(ds.label) == 10
        ds.checkout("main")
        with pytest.raises(MergeNotSupportedError):
            ds.merge("other")


def test_tensor_mismatch(local_ds):
    with local_ds as ds:
        ds.create_tensor("image")
        ds.checkout("alt", create=True)
        ds.create_tensor("xyz", htype="bbox")
        ds.checkout("main")
        ds.create_tensor("xyz", htype="class_label")
        with pytest.raises(MergeMismatchError):
            ds.merge("alt")


def test_new_tensor_creation_merge(local_ds):
    with local_ds as ds:
        ds.create_tensor("image")
        ds.checkout("alt", create=True)
        ds.create_tensor("xyz")
        for i in range(100):
            ds.xyz.append(i)
        ds.commit()
        ds.checkout("main")
        ds.merge("alt")
        assert "xyz" in ds.tensors
        assert len(ds.xyz) == 100
        for i in range(100):
            assert ds.xyz[i].numpy() == i

        # merging twice to confirm, that extra items are not added
        ds.merge("alt")
        assert "xyz" in ds.tensors
        assert len(ds.xyz) == 100
        for i in range(100):
            assert ds.xyz[i].numpy() == i


def test_tensor_deletion_merge(local_ds):
    with local_ds as ds:
        ds.create_tensor("image")
        ds.create_tensor("label")
        for i in range(10):
            ds.image.append(i * np.ones((200, 200, 3)))
            ds.label.append(i)

        a = ds.commit("added 10 images and labels")
        ds.checkout("other", create=True)

        ds.delete_tensor("image")
        ds.label.append(10)
        ds.commit()
        ds.checkout("main")
        ds.merge("other")
        assert "image" in ds.tensors
        for i in range(10):
            np.testing.assert_array_equal(
                ds.image[i].numpy(), i * np.ones((200, 200, 3))
            )
            assert ds.label[i].numpy() == i
        assert ds.label[10].numpy() == 10

        ds.merge("other", delete_removed_tensors=True)
        assert "image" not in ds.tensors
        assert "label" in ds.tensors
        for i in range(11):
            assert ds.label[i].numpy() == i


def test_tensor_revival(local_ds):
    with local_ds as ds:
        ds.create_tensor("image")
        ds.create_tensor("label")

        for i in range(10):
            ds.image.append(i * np.ones((200, 200, 3)))
            ds.label.append(i)

        a = ds.commit("added 10 images and labels")
        ds.delete_tensor("image")
        ds.delete_tensor("label")

        ds.checkout(a)
        ds.checkout("other", create=True)

        assert "image" in ds.tensors
        assert "label" in ds.tensors
        for i in range(10, 15):
            ds.label.append(i)
        ds.commit()
        ds.checkout("main")
        assert "image" not in ds.tensors
        assert "label" not in ds.tensors

        ds.merge("other")

        # not revived as this had no changes in other
        assert "image" not in ds.tensors

        # revived as this had changes in other
        assert "label" in ds.tensors
        for i in range(15):
            assert ds.label[i].numpy() == i


def test_conflicts(local_ds):
    with local_ds as ds:
        ds.create_tensor("image")
        for i in range(10):
            ds.image.append(i * np.ones((200, 200, 3)))

        a = ds.commit("added 10 images")
        ds.checkout("other", create=True)
        for i in range(10, 15):
            ds.image.append(i * np.ones((200, 200, 3)))
        ds.image[4] = 25 * np.ones((200, 200, 3))

        b = ds.commit("added 5 more images and changed 4th")
        ds.checkout("main")
        ds.image[4] = 50 * np.ones((200, 200, 3))

        with pytest.raises(MergeConflictError):
            ds.merge("other")

        ds.merge("other", conflict_resolution="theirs")
        np.testing.assert_array_equal(ds.image[4].numpy(), 25 * np.ones((200, 200, 3)))


def test_rename_merge(local_ds):
    with local_ds as ds:
        # no conflicts
        ds.create_tensor("abc")
        ds.abc.append([1, 2, 3])
        ds.commit()
        ds.checkout("alt", create=True)
        ds.rename_tensor("abc", "xyz")
        ds.xyz.append([3, 4, 5])
        ds.commit()
        ds.checkout("main")
        ds.merge("alt")
        assert "abc" not in ds.tensors
        np.testing.assert_array_equal(ds.xyz.numpy(), np.array([[1, 2, 3], [3, 4, 5]]))

        # tensor with same name on main
        ds.create_tensor("red")
        ds.red.append([2, 3, 4])
        ds.commit()
        ds.checkout("alt2", create=True)
        ds.rename_tensor("red", "blue")
        ds.blue.append([1, 0, 0])
        ds.commit()
        ds.checkout("main")
        ds.create_tensor("blue")
        ds.blue.append([0, 0, 1])
        ds.commit()
        with pytest.raises(MergeConflictError):
            ds.merge("alt2")
        # resolve
        ds.merge("alt2", force=True)
        np.testing.assert_array_equal(ds.red.numpy(), np.array([[2, 3, 4]]))
        np.testing.assert_array_equal(
            ds.blue.numpy(), np.array([[0, 0, 1], [2, 3, 4], [1, 0, 0]])
        )
        ds.delete_tensor("blue")
        ds.commit()

        # rename to same name (no conflict)
        ds.rename_tensor("red", "blue")
        ds.merge("alt2")
        np.testing.assert_array_equal(ds.blue.numpy(), np.array([[2, 3, 4], [1, 0, 0]]))

        # renamed on both branches
        ds.create_tensor("image")
        ds.create_tensor("video")
        ds.image.append([1, 2, 3])
        ds.video.append([1, 0, 0])
        ds.commit()
        ds.checkout("alt3", create=True)
        ds.rename_tensor("image", "images")
        ds.rename_tensor("video", "videos")
        ds.images.append([3, 4, 5])
        ds.videos.append([0, 1, 0])
        ds.commit()
        ds.checkout("main")
        ds.rename_tensor("image", "cat_images")
        ds.rename_tensor("video", "sample_videos")
        ds.cat_images.append([5, 6, 7])
        ds.sample_videos.append([7, 8, 9])
        ds.commit()
        with pytest.raises(MergeConflictError):
            ds.merge("alt3")
        # resolve (through rename and force)
        ds.checkout("alt3")
        ds.rename_tensor("images", "cat_images")
        ds.commit()
        ds.checkout("main")
        ds.merge("alt3", force=True)
        assert "image" not in ds.tensors
        assert "images" not in ds.tensors
        np.testing.assert_array_equal(
            ds.cat_images.numpy(), np.array([[1, 2, 3], [5, 6, 7], [3, 4, 5]])
        )
        np.testing.assert_array_equal(
            ds.videos.numpy(), np.array([[1, 0, 0], [0, 1, 0]])
        )
        np.testing.assert_array_equal(
            ds.sample_videos.numpy(), np.array([[1, 0, 0], [7, 8, 9]])
        )

        # tensor missing from head
        ds.create_tensor("t1")
        ds.t1.append([1, 2, 3])
        ds.commit()
        ds.checkout("alt4", create=True)
        ds.rename_tensor("t1", "t2")
        ds.t2.append([4, 5, 6])
        ds.commit()
        ds.checkout("main")
        ds.delete_tensor("t1")
        ds.commit()
        with pytest.raises(MergeConflictError):
            ds.merge("alt4")


def test_clear_merge(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc")
        ds.abc.append([1, 2, 3])
        a = ds.commit()

        ds.checkout("alt", create=True)
        ds.abc.append([2, 3, 4])
        b = ds.commit()
        ds.abc.clear()
        c = ds.commit()

        ds.checkout("main")
        ds.abc.append([5, 6, 3])
        d = ds.commit()
        ds.merge("alt")

        np.testing.assert_array_equal(ds.abc.numpy(), np.array([]))



import hub
import pytest
import numpy as np
from click.testing import CliRunner
from hub.core.storage.memory import MemoryProvider
from hub.util.remove_cache import remove_memory_cache
from hub.util.check_installation import ray_installed
from hub.util.exceptions import InvalidOutputDatasetError, TransformError
from hub.tests.common import parametrize_num_workers
from hub.util.transform import get_pbar_description
import hub


# github actions can only support 2 workers
TRANSFORM_TEST_NUM_WORKERS = 2

all_compressions = pytest.mark.parametrize("sample_compression", [None, "png", "jpeg"])

schedulers = ["threaded", "processed"]
schedulers = schedulers + ["ray"] if ray_installed() else schedulers
all_schedulers = pytest.mark.parametrize("scheduler", schedulers)
commit_or_not = pytest.mark.parametrize("do_commit", [True, False])


@hub.compute
def fn1(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.image.append(np.ones((337, 200)) * sample_in * mul)
        samples_out.label.append(np.ones((1,)) * sample_in * mul)


@hub.compute
def fn2(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.image.append(sample_in.image.numpy() * mul)
        samples_out.label.append(sample_in.label.numpy() * mul)


@hub.compute
def fn3(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.image.append(np.ones((1310, 2087)) * sample_in * mul)
        samples_out.label.append(np.ones((13,)) * sample_in * mul)


@hub.compute
def fn4(sample_in, samples_out):
    samples_out.image.append(sample_in.image)
    samples_out.image.append(sample_in.image.numpy() * 2)
    samples_out.label.append(sample_in.label)
    samples_out.label.append(sample_in.label.numpy() * 2)


@hub.compute
def fn5(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.x["y"].z.image.append(sample_in.z.y.x.image.numpy() * mul)
        samples_out.x.y.z["label"].append(sample_in.z.y.x.label.numpy() * mul)


@hub.compute
def fn6(sample_in, samples_out, mul=1, copy=1):
    for _ in range(copy):
        samples_out.append(
            {
                "image": sample_in.image.numpy() * mul,
                "label": sample_in.label.numpy() * mul,
            }
        )


@hub.compute
def read_image(sample_in, samples_out):
    samples_out.image.append(hub.read(sample_in))


@hub.compute
def crop_image(sample_in, samples_out, copy=1):
    for _ in range(copy):
        samples_out.image.append(sample_in.image.numpy()[:100, :100, :])


@hub.compute
def filter_tr(sample_in, sample_out):
    if sample_in % 2 == 0:
        sample_out.image.append(sample_in * np.ones((100, 100)))


@hub.compute
def populate_cc_bug(sample_in, samples_out):
    samples_out.xyz.append(sample_in)
    return samples_out


@hub.compute
def inplace_transform(sample_in, samples_out):
    samples_out.img.append(2 * sample_in.img.numpy())
    samples_out.img.append(3 * sample_in.img.numpy())
    samples_out.label.append(2 * sample_in.label.numpy())
    samples_out.label.append(3 * sample_in.label.numpy())


@hub.compute
def unequal_transform(sample_in, samples_out):
    samples_out.x.append(sample_in.x.numpy() * 2)
    if sample_in.y.numpy().size > 0:
        samples_out.y.append(sample_in.y.numpy() * 2)


@hub.compute
def add_text(sample_in, samples_out):
    samples_out.abc.append(sample_in)


@hub.compute
def add_link(sample_in, samples_out):
    samples_out.abc.append(hub.link(sample_in))


def check_target_array(ds, index, target):
    np.testing.assert_array_equal(
        ds.img[index].numpy(), target * np.ones((200, 200, 3))
    )
    np.testing.assert_array_equal(ds.label[index].numpy(), target * np.ones((1,)))


@all_schedulers
@pytest.mark.parametrize(
    "ds",
    ["memory_ds", "local_ds", "s3_ds"],
    indirect=True,
)
def test_single_transform_hub_dataset(ds, scheduler):
    data_in = hub.dataset("./test/single_transform_hub_dataset", overwrite=True)
    with data_in:
        data_in.create_tensor("image")
        data_in.create_tensor("label")
        for i in range(1, 100):
            data_in.image.append(i * np.ones((i, i)))
            data_in.label.append(i * np.ones((1,)))
    ds_out = ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    if (
        isinstance(remove_memory_cache(ds.storage), MemoryProvider)
        and scheduler != "threaded"
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        with pytest.raises(InvalidOutputDatasetError):
            fn2(copy=1, mul=2).eval(
                data_in,
                ds_out,
                num_workers=TRANSFORM_TEST_NUM_WORKERS,
                progressbar=False,
                scheduler=scheduler,
            )
        data_in.delete()
        return

    fn2(copy=1, mul=2).eval(
        data_in,
        ds_out,
        num_workers=TRANSFORM_TEST_NUM_WORKERS,
        scheduler=scheduler,
        progressbar=False,
    )
    assert len(ds_out) == 99
    for index in range(1, 100):
        np.testing.assert_array_equal(
            ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
        )
        np.testing.assert_array_equal(
            ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
        )

    assert ds_out.image.shape_interval.lower == (99, 1, 1)
    assert ds_out.image.shape_interval.upper == (99, 99, 99)
    data_in.delete()


def test_groups(local_ds):
    with CliRunner().isolated_filesystem():
        with hub.dataset("./test/transform_hub_in_generic") as data_in:
            data_in.create_tensor("data/image")
            data_in.create_tensor("data/label")
            for i in range(1, 100):
                data_in.data.image.append(i * np.ones((i, i)))
                data_in.data.label.append(i * np.ones((1,)))
        data_in = hub.dataset("./test/transform_hub_in_generic")
        ds_out = local_ds
        ds_out.create_tensor("stuff/image")
        ds_out.create_tensor("stuff/label")

        data_in = data_in.data
        ds_out = ds_out.stuff

        fn2(copy=1, mul=2).eval(
            data_in, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, progressbar=False
        )
        assert len(ds_out) == 99
        for index in range(1, 100):
            np.testing.assert_array_equal(
                ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
            )
            np.testing.assert_array_equal(
                ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
            )

        assert ds_out.image.shape_interval.lower == (99, 1, 1)
        assert ds_out.image.shape_interval.upper == (99, 99, 99)


def test_groups_2(local_ds):
    with CliRunner().isolated_filesystem():
        with hub.dataset("./test/transform_hub_in_generic") as data_in:
            data_in.create_tensor("data/z/y/x/image")
            data_in.create_tensor("data/z/y/x/label")
            for i in range(1, 100):
                data_in.data.z.y.x.image.append(i * np.ones((i, i)))
                data_in.data.z.y.x.label.append(i * np.ones((1,)))
        data_in = hub.dataset("./test/transform_hub_in_generic")
        ds_out = local_ds
        ds_out.create_tensor("stuff/x/y/z/image")
        ds_out.create_tensor("stuff/x/y/z/label")

        data_in = data_in.data
        ds_out = ds_out.stuff

        fn5(copy=1, mul=2).eval(
            data_in, ds_out, num_workers=TRANSFORM_TEST_NUM_WORKERS, progressbar=False
        )
        assert len(ds_out) == 99
        for index in range(1, 100):
            np.testing.assert_array_equal(
                ds_out.x.y.z.image[index - 1].numpy(),
                2 * index * np.ones((index, index)),
            )
            np.testing.assert_array_equal(
                ds_out.x.y.z.label[index - 1].numpy(), 2 * index * np.ones((1,))
            )

        assert ds_out.x.y.z.image.shape_interval.lower == (99, 1, 1)
        assert ds_out.x.y.z.image.shape_interval.upper == (99, 99, 99)


@parametrize_num_workers
@all_schedulers
def test_single_transform_hub_dataset_htypes(local_ds, num_workers, scheduler):
    data_in = hub.dataset("./test/single_transform_hub_dataset_htypes", overwrite=True)
    with data_in:
        data_in.create_tensor("image", htype="image", sample_compression="png")
        data_in.create_tensor("label", htype="class_label")
        for i in range(1, 100):
            data_in.image.append(i * np.ones((i, i), dtype="uint8"))
            data_in.label.append(i * np.ones((1,), dtype="uint32"))
    ds_out = local_ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    fn2(copy=1, mul=2).eval(
        data_in, ds_out, num_workers=num_workers, progressbar=False, scheduler=scheduler
    )
    assert len(ds_out) == 99
    for index in range(1, 100):
        np.testing.assert_array_equal(
            ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
        )
        np.testing.assert_array_equal(
            ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
        )

    assert ds_out.image.shape_interval.lower == (99, 1, 1)
    assert ds_out.image.shape_interval.upper == (99, 99, 99)
    data_in.delete()


@all_schedulers
def test_chain_transform_list_small(local_ds, scheduler):
    ls = list(range(100))
    ds_out = local_ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    pipeline = hub.compose([fn1(mul=5, copy=2), fn2(mul=3, copy=3)])
    pipeline.eval(
        ls,
        ds_out,
        num_workers=TRANSFORM_TEST_NUM_WORKERS,
        progressbar=False,
        scheduler=scheduler,
    )
    assert len(ds_out) == 600
    for i in range(100):
        for index in range(6 * i, 6 * i + 6):
            np.testing.assert_array_equal(
                ds_out[index].image.numpy(), 15 * i * np.ones((337, 200))
            )
            np.testing.assert_array_equal(
                ds_out[index].label.numpy(), 15 * i * np.ones((1,))
            )


@all_schedulers
def test_chain_transform_list_big(local_ds, scheduler):
    ls = [i for i in range(2)]
    ds_out = local_ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    pipeline = hub.compose([fn3(mul=5, copy=2), fn2(mul=3, copy=3)])
    pipeline.eval(
        ls,
        ds_out,
        num_workers=TRANSFORM_TEST_NUM_WORKERS,
        progressbar=False,
        scheduler=scheduler,
    )
    assert len(ds_out) == 12
    for i in range(2):
        for index in range(6 * i, 6 * i + 6):
            np.testing.assert_array_equal(
                ds_out[index].image.numpy(), 15 * i * np.ones((1310, 2087))
            )
            np.testing.assert_array_equal(
                ds_out[index].label.numpy(), 15 * i * np.ones((13,))
            )


@all_schedulers
@commit_or_not
def test_add_to_non_empty_dataset(local_ds, scheduler, do_commit):
    ls = [i for i in range(100)]
    ds_out = local_ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    pipeline = hub.compose([fn1(mul=5, copy=2), fn2(mul=3, copy=3)])
    with ds_out:
        for i in range(10):
            ds_out.image.append(i * np.ones((10, 10)))
            ds_out.label.append(i * np.ones((1,)))
        if do_commit:
            ds_out.commit()

    pipeline.eval(
        ls,
        ds_out,
        num_workers=TRANSFORM_TEST_NUM_WORKERS,
        progressbar=False,
        scheduler=scheduler,
    )
    assert len(ds_out) == 610
    for i in range(10):
        np.testing.assert_array_equal(ds_out[i].image.numpy(), i * np.ones((10, 10)))
        np.testing.assert_array_equal(ds_out[i].label.numpy(), i * np.ones((1,)))
    for i in range(100):
        for index in range(10 + 6 * i, 10 + 6 * i + 6):
            np.testing.assert_array_equal(
                ds_out[index].image.numpy(), 15 * i * np.ones((337, 200))
            )
            np.testing.assert_array_equal(
                ds_out[index].label.numpy(), 15 * i * np.ones((1,))
            )

    diff = ds_out.diff(as_dict=True)["tensor"]
    change = {
        "image": {
            "data_updated": set(),
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
        "label": {
            "data_updated": set(),
            "info_updated": False,
            "data_transformed_in_place": False,
            "data_deleted": set(),
        },
    }
    if do_commit:
        change["image"]["cleared"] = False
        change["label"]["cleared"] = False
        change["image"]["created"] = False
        change["label"]["created"] = False
        change["image"]["data_added"] = [10, 610]
        change["label"]["data_added"] = [10, 610]
    else:
        change["image"]["cleared"] = False
        change["label"]["cleared"] = False
        change["image"]["created"] = True
        change["label"]["created"] = True
        change["image"]["data_added"] = [0, 610]
        change["label"]["data_added"] = [0, 610]
    assert diff == change


@all_schedulers
@all_compressions
def test_transform_hub_read(local_ds, cat_path, sample_compression, scheduler):
    data_in = [cat_path] * 10
    ds_out = local_ds
    ds_out.create_tensor("image", htype="image", sample_compression=sample_compression)

    read_image().eval(
        data_in,
        ds_out,
        num_workers=TRANSFORM_TEST_NUM_WORKERS,
        progressbar=False,
        scheduler=scheduler,
    )
    assert len(ds_out) == 10
    for i in range(10):
        assert ds_out.image[i].numpy().shape == (900, 900, 3)
        np.testing.assert_array_equal(ds_out.image[i].numpy(), ds_out.image[0].numpy())


@all_schedulers
@all_compressions
def test_transform_hub_read_pipeline(local_ds, cat_path, sample_compression, scheduler):
    data_in = [cat_path] * 10
    ds_out = local_ds
    ds_out.create_tensor("image", htype="image", sample_compression=sample_compression)
    pipeline = hub.compose([read_image(), crop_image(copy=2)])
    pipeline.eval(
        data_in,
        ds_out,
        num_workers=TRANSFORM_TEST_NUM_WORKERS,
        progressbar=False,
        scheduler=scheduler,
    )
    assert len(ds_out) == 20
    for i in range(20):
        assert ds_out.image[i].numpy().shape == (100, 100, 3)
        np.testing.assert_array_equal(ds_out.image[i].numpy(), ds_out.image[0].numpy())


def test_hub_like(local_ds, scheduler="threaded"):
    with CliRunner().isolated_filesystem():
        data_in = local_ds
        with data_in:
            data_in.create_tensor("image", htype="image", sample_compression="png")
            data_in.create_tensor("label", htype="class_label")
            for i in range(1, 100):
                data_in.image.append(i * np.ones((i, i), dtype="uint8"))
                data_in.label.append(i * np.ones((1,), dtype="uint32"))
        ds_out = hub.like("./transform_hub_like", data_in)
        fn2(copy=1, mul=2).eval(
            data_in,
            ds_out,
            num_workers=TRANSFORM_TEST_NUM_WORKERS,
            progressbar=False,
            scheduler=scheduler,
        )
        assert len(ds_out) == 99
        for index in range(1, 100):
            np.testing.assert_array_equal(
                ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
            )
            np.testing.assert_array_equal(
                ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
            )

        assert ds_out.image.shape_interval.lower == (99, 1, 1)
        assert ds_out.image.shape_interval.upper == (99, 99, 99)


def test_transform_empty(local_ds):
    local_ds.create_tensor("image")

    ls = list(range(10))
    filter_tr().eval(ls, local_ds, progressbar=False)

    assert len(local_ds) == 5

    for i in range(5):
        np.testing.assert_array_equal(
            local_ds[i].image.numpy(), 2 * i * np.ones((100, 100))
        )


def test_pbar_description():
    assert get_pbar_description([fn1()]) == "Evaluating fn1"
    assert get_pbar_description([fn1(), fn2()]) == "Evaluating [fn1, fn2]"
    assert get_pbar_description([fn1(), fn1()]) == "Evaluating [fn1, fn1]"
    assert (
        get_pbar_description([fn1(), fn1(), read_image()])
        == "Evaluating [fn1, fn1, read_image]"
    )


def test_bad_transform(memory_ds):
    ds = memory_ds
    ds.create_tensor("x")
    ds.create_tensor("y")
    with ds:
        ds.x.extend(np.random.rand(10, 1))
        ds.y.extend(np.random.rand(10, 1))
    ds2 = hub.like("mem://dummy2", ds)

    @hub.compute
    def fn_filter(sample_in, sample_out):
        sample_out.y.append(sample_in.y.numpy())
        return sample_out

    with pytest.raises(TransformError):
        fn_filter().eval(ds, ds2, progressbar=False)


def test_transform_persistance(local_ds_generator, num_workers=2, scheduler="threaded"):
    data_in = hub.dataset("./test/single_transform_hub_dataset_htypes", overwrite=True)
    with data_in:
        data_in.create_tensor("image", htype="image", sample_compression="png")
        data_in.create_tensor("label", htype="class_label")
        for i in range(1, 100):
            data_in.image.append(i * np.ones((i, i), dtype="uint8"))
            data_in.label.append(i * np.ones((1,), dtype="uint32"))
    ds_out = local_ds_generator()
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")
    if (
        isinstance(remove_memory_cache(ds_out.storage), MemoryProvider)
        and scheduler != "threaded"
        and num_workers > 0
    ):
        # any scheduler other than `threaded` will not work with a dataset stored in memory
        # num_workers = 0 automatically does single threaded irrespective of the scheduler
        with pytest.raises(InvalidOutputDatasetError):
            fn2(copy=1, mul=2).eval(
                data_in,
                ds_out,
                num_workers=num_workers,
                scheduler=scheduler,
                progressbar=False,
            )
        data_in.delete()
        return
    fn2(copy=1, mul=2).eval(
        data_in, ds_out, num_workers=num_workers, scheduler=scheduler, progressbar=False
    )

    def test_ds_out():
        assert len(ds_out) == 99
        for index in range(1, 100):
            np.testing.assert_array_equal(
                ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
            )
            np.testing.assert_array_equal(
                ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
            )

        assert ds_out.image.shape_interval.lower == (99, 1, 1)
        assert ds_out.image.shape_interval.upper == (99, 99, 99)

    test_ds_out()
    ds_out = local_ds_generator()
    test_ds_out()

    data_in.delete()


def test_ds_append_in_transform(memory_ds):
    ds = memory_ds
    data_in = hub.dataset("./test/single_transform_hub_dataset", overwrite=True)
    with data_in:
        data_in.create_tensor("image")
        data_in.create_tensor("label")
        for i in range(1, 100):
            data_in.image.append(i * np.ones((i, i)))
            data_in.label.append(i * np.ones((1,)))
    ds_out = ds
    ds_out.create_tensor("image")
    ds_out.create_tensor("label")

    fn6(copy=1, mul=2).eval(
        data_in, ds_out, num_workers=2, scheduler="threaded", progressbar=False
    )
    assert len(ds_out) == 99
    for index in range(1, 100):
        np.testing.assert_array_equal(
            ds_out[index - 1].image.numpy(), 2 * index * np.ones((index, index))
        )
        np.testing.assert_array_equal(
            ds_out[index - 1].label.numpy(), 2 * index * np.ones((1,))
        )

    assert ds_out.image.shape_interval.lower == (99, 1, 1)
    assert ds_out.image.shape_interval.upper == (99, 99, 99)
    data_in.delete()


def test_transform_pass_through():
    data_in = hub.dataset("mem://ds1")
    data_in.create_tensor("image", htype="image", sample_compression="png")
    data_in.create_tensor("label", htype="class_label")
    for i in range(1, 100):
        data_in.image.append(i * np.ones((i, i), dtype="uint8"))
        data_in.label.append(i * np.ones((1,), dtype="uint32"))
    ds_out = hub.dataset("mem://ds2")
    ds_out.create_tensor("image", htype="image", sample_compression="png")
    ds_out.create_tensor("label", htype="class_label")
    fn4().eval(data_in, ds_out, num_workers=2, scheduler="threaded", progressbar=False)
    for i in range(len(data_in)):
        np.testing.assert_array_equal(
            data_in[i].image.numpy(), ds_out[i * 2].image.numpy()
        )
        np.testing.assert_array_equal(
            data_in[i].label.numpy(), ds_out[i * 2].label.numpy()
        )
        np.testing.assert_array_equal(
            data_in[i].image.numpy() * 2, ds_out[i * 2 + 1].image.numpy()
        )
        np.testing.assert_array_equal(
            data_in[i].label.numpy() * 2, ds_out[i * 2 + 1].label.numpy()
        )


def test_inplace_transform(local_ds_generator):
    ds = local_ds_generator()

    with ds:
        ds.create_tensor("img")
        ds.create_tensor("label")
        for i in range(10):
            if i == 5:
                ds.img.append(np.zeros((200, 200, 3)))
            else:
                ds.img.append(np.ones((200, 200, 3)))
            ds.label.append(1)
        a = ds.commit()
        assert len(ds) == 10
        for i in range(10):
            if i != 5:
                check_target_array(ds, i, 1)
        ds.img[5] = np.ones((200, 200, 3))
        b = ds.commit()

        inplace_transform().eval(
            ds, num_workers=TRANSFORM_TEST_NUM_WORKERS, progressbar=False
        )
        assert ds.img.chunk_engine.num_samples == len(ds) == 20

        for i in range(20):
            target = 2 if i % 2 == 0 else 3
            check_target_array(ds, i, target)

        diff = ds.diff(as_dict=True)["tensor"]
        change = {
            "img": {
                "created": False,
                "cleared": False,
                "data_added": [0, 20],
                "data_updated": set(),
                "data_transformed_in_place": True,
                "info_updated": False,
                "data_deleted": set(),
            },
            "label": {
                "created": False,
                "cleared": False,
                "data_added": [0, 20],
                "data_updated": set(),
                "data_transformed_in_place": True,
                "info_updated": False,
                "data_deleted": set(),
            },
        }
        assert diff == change

        ds.checkout(b)
        assert len(ds) == 10
        for i in range(10):
            check_target_array(ds, i, 1)

    ds = local_ds_generator()
    assert len(ds) == 20
    for i in range(20):
        target = 2 if i % 2 == 0 else 3
        check_target_array(ds, i, target)

    ds.checkout(b)
    assert len(ds) == 10
    for i in range(10):
        check_target_array(ds, i, 1)


def test_inplace_transform_without_commit(local_ds_generator):
    ds = local_ds_generator()

    with ds:
        ds.create_tensor("img")
        ds.create_tensor("label")
        for _ in range(10):
            ds.img.append(np.ones((200, 200, 3)))
            ds.label.append(1)
        assert len(ds) == 10
        for i in range(10):
            check_target_array(ds, i, 1)

        inplace_transform().eval(
            ds, num_workers=TRANSFORM_TEST_NUM_WORKERS, progressbar=False
        )
        assert ds.img.chunk_engine.num_samples == len(ds) == 20

        for i in range(20):
            target = 2 if i % 2 == 0 else 3
            check_target_array(ds, i, target)

    ds = local_ds_generator()
    assert len(ds) == 20
    for i in range(20):
        target = 2 if i % 2 == 0 else 3
        check_target_array(ds, i, target)


def test_inplace_transform_non_head(local_ds_generator):
    ds = local_ds_generator()
    with ds:
        ds.create_tensor("img")
        ds.create_tensor("label")
        for _ in range(10):
            ds.img.append(np.ones((200, 200, 3)))
            ds.label.append(1)
        assert len(ds) == 10
        for i in range(10):
            check_target_array(ds, i, 1)
        a = ds.commit()
        for _ in range(5):
            ds.img.append(np.ones((200, 200, 3)))
            ds.label.append(1)
        assert len(ds) == 15
        for i in range(15):
            check_target_array(ds, i, 1)

        ds.checkout(a)

        # transforming non-head node
        inplace_transform().eval(
            ds, num_workers=TRANSFORM_TEST_NUM_WORKERS, progressbar=False
        )
        br = ds.branch

        assert len(ds) == 20
        for i in range(20):
            target = 2 if i % 2 == 0 else 3
            check_target_array(ds, i, target)

        ds.checkout(a)
        assert len(ds) == 10
        for i in range(10):
            check_target_array(ds, i, 1)

        ds.checkout("main")
        assert len(ds) == 15
        for i in range(15):
            check_target_array(ds, i, 1)

    ds = local_ds_generator()
    assert len(ds) == 15
    for i in range(15):
        check_target_array(ds, i, 1)

    ds.checkout(a)
    assert len(ds) == 10
    for i in range(10):
        check_target_array(ds, i, 1)

    ds.checkout(br)
    assert len(ds) == 20
    for i in range(20):
        target = 2 if i % 2 == 0 else 3
        check_target_array(ds, i, target)


def test_inplace_transform_clear_chunks(local_ds_generator):
    ds = local_ds_generator()

    with ds:
        ds.create_tensor("img")
        ds.create_tensor("label")

        for _ in range(10):
            ds.img.append(np.ones((500, 500, 3)))
            ds.label.append(np.ones(3))

    prev_chunks = set(
        [
            f"{tensor.key}/chunks/{chunk}"
            for tensor in [ds.img, ds.label]
            for chunk in tensor.chunk_engine.list_all_chunks()
        ]
    )
    inplace_transform().eval(ds)
    after_chunks = set(
        [
            f"{tensor.key}/chunks/{chunk}"
            for tensor in [ds.img, ds.label]
            for chunk in tensor.chunk_engine.list_all_chunks()
        ]
    )

    # all chunks where replaced
    assert len(after_chunks.intersection(prev_chunks)) == 0

    # test all new chunks where created
    for chunk in after_chunks:
        assert ds.storage[chunk] is not None

    # test all old chunks where removed
    for chunk in prev_chunks:
        try:
            assert ds.storage[chunk] is None
        except KeyError:
            pass


def test_transform_skip_ok(local_ds_generator):
    ds = local_ds_generator()
    ls = list(range(100))
    with ds:
        ds.create_tensor("image")
        ds.create_tensor("label")
        ds.create_tensor("unused")

    pipeline = hub.compose([fn1(mul=5, copy=2), fn2(mul=3, copy=3)])
    pipeline.eval(
        ls,
        ds,
        num_workers=TRANSFORM_TEST_NUM_WORKERS,
        progressbar=False,
        scheduler="processed",
        skip_ok=True,
    )
    for i in range(100):
        for index in range(6 * i, 6 * i + 6):
            np.testing.assert_array_equal(
                ds.image[index].numpy(), 15 * i * np.ones((337, 200))
            )
            np.testing.assert_array_equal(
                ds.label[index].numpy(), 15 * i * np.ones((1,))
            )

    assert len(ds.unused) == 0

    # test persistence
    ds = local_ds_generator()
    for i in range(100):
        for index in range(6 * i, 6 * i + 6):
            np.testing.assert_array_equal(
                ds.image[index].numpy(), 15 * i * np.ones((337, 200))
            )
            np.testing.assert_array_equal(
                ds.label[index].numpy(), 15 * i * np.ones((1,))
            )
    assert len(ds.unused) == 0


def test_inplace_transform_skip_ok(local_ds_generator):
    ds = local_ds_generator()

    with ds:
        ds.create_tensor("img")
        ds.create_tensor("label")
        ds.create_tensor("unused")
        ds.img.extend(np.ones((10, 200, 200, 3)))
        ds.label.extend([1 for _ in range(10)])
        ds.unused.extend(5 * np.ones((10, 10, 10)))
        for i in range(10):
            check_target_array(ds, i, 1)

    inplace_transform().eval(
        ds, num_workers=TRANSFORM_TEST_NUM_WORKERS, progressbar=False, skip_ok=True
    )
    assert ds.img.chunk_engine.num_samples == 20

    for i in range(20):
        target = 2 if i % 2 == 0 else 3
        check_target_array(ds, i, target)

    assert len(ds.unused) == 10
    np.testing.assert_array_equal(ds.unused.numpy(), 5 * np.ones((10, 10, 10)))

    # test persistence
    ds = local_ds_generator()

    assert ds.img.chunk_engine.num_samples == 20

    for i in range(20):
        target = 2 if i % 2 == 0 else 3
        check_target_array(ds, i, target)

    assert len(ds.unused) == 10
    np.testing.assert_array_equal(ds.unused.numpy(), 5 * np.ones((10, 10, 10)))


def test_chunk_compression_bug(local_ds):
    xyz = np.zeros((480, 640), dtype=np.float32)
    length = 55
    dataset = [xyz] * length
    with local_ds as ds:
        ds.create_tensor("xyz", chunk_compression="lz4")
        populate_cc_bug().eval(dataset, ds, num_workers=2, scheduler="threaded")

    for index in range(length):
        np.testing.assert_array_equal(ds.xyz[index].numpy(), xyz)


@hub.compute
def sequence_transform(inp, out):
    out.x.append([np.ones(inp)] * inp)


def test_sequence_htype_with_transform(local_ds):
    ds = local_ds
    with ds:
        ds.create_tensor("x", htype="sequence")
        assert ds.x.dtype is None
        assert ds.x.htype == "sequence[None]"
        sequence_transform().eval(list(range(1, 11)), ds, TRANSFORM_TEST_NUM_WORKERS)
    for i in range(10):
        np.testing.assert_array_equal(ds.x[i].numpy(), np.ones((i + 1, i + 1)))
    assert ds.x.dtype == np.ones(1).dtype
    assert ds.x.htype == "sequence[generic]"


def test_htype_dtype_after_transform(local_ds):
    ds = local_ds
    with ds:
        ds.create_tensor("image")
        assert ds.image.htype is None
        assert ds.image.dtype is None
        ds.create_tensor("label")
        fn3().eval(list(range(10)), ds, TRANSFORM_TEST_NUM_WORKERS)
    assert ds.image.htype == "generic"
    assert ds.image.dtype == np.ones(1).dtype


def test_transform_pad_data_in(local_ds):
    with local_ds as ds:
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds.x.extend(list(range(10)))
        ds.y.extend(list(range(5)))
    ds2 = hub.dataset("./data/unequal2", overwrite=True)
    ds2.create_tensor("x")
    ds2.create_tensor("y")

    unequal_transform().eval(ds, ds2, pad_data_in=True, skip_ok=True)
    assert len(ds2.x) == 10
    assert len(ds2.y) == 5
    assert len(ds2) == 5
    for i in range(10):
        x = ds2[i].x.numpy()
        np.testing.assert_equal(x, 2 * i)
        if i < 5:
            y = ds2[i].y.numpy()
            np.testing.assert_equal(y, 2 * i)

    for i, dsv in enumerate(ds2):
        x, y = dsv.x.numpy(), dsv.y.numpy()
        np.testing.assert_equal(x, 2 * i)
        np.testing.assert_equal(y, 2 * i)


def test_transform_bug_text(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc", htype="text")
        ls = ["hello"] * 10
        add_text().eval(ls, ds, num_workers=2)
        assert len(ds) == 10
        ds.pop(6)
        assert len(ds) == 9

        for i in range(9):
            assert ds[i].abc.numpy() == "hello"


def test_transform_bug_link(local_ds, cat_path):
    with local_ds as ds:
        ds.create_tensor("abc", htype="link[image]")
        ls = [cat_path] * 10
        add_link().eval(ls, ds, num_workers=2)
        assert len(ds) == 10
        ds.pop(6)
        assert len(ds) == 9

        for i in range(9):
            assert ds[i].abc.numpy().shape == (900, 900, 3)
            assert ds[i].abc.shape == (900, 900, 3)



from typing import Union

import tensorflow  # type: ignore
import tensorflow_datasets as tfds  # type: ignore
from tensorflow_datasets import Split
from tqdm import tqdm  # type: ignore

from hub.core.dataset import Dataset
import hub


def from_tfds_to_path(
    tfds_dataset_name: str,
    split: Union[str, Split],
    hub_ds_path: str,
    batch_size: int = 100,
):
    """Converts the tfds dataset with name `tfds_dataset_name` into a hub dataset and saves it at `hub_ds_path`
    Args:
        tfds_dataset_name (str): Name of tfds dataset.You can see a list of all tfds datasets here:
            https://www.tensorflow.org/datasets/catalog/overview
        split (str, Split) : Used for dataset splits as defined here: https://www.tensorflow.org/datasets/splits
        hub_ds_path (str): Path where new hub dataset will be created
        batch_size (int): Batch size for tfds dataset. Has no effect on output, but may affect performance.
    Returns:
        A hub dataset
    """
    tfds_ds = tfds.load(tfds_dataset_name, split=split).batch(batch_size)
    ds = hub.dataset(hub_ds_path)

    return from_tfds(tfds_ds=tfds_ds, ds=ds)  # type: ignore


def from_tfds(tfds_ds: tensorflow.data.Dataset, ds: Dataset):
    """Converts a tfds dataset to hub dataset
    Args:
        tfds_ds (tensorflow.data.Dataset): A tfds_dataset object.
        ds (Dataset) : A Hub dataset object where Tensor will be created.
    Returns:
        A hub dataset
    """
    tfds_numpy = tfds.as_numpy(tfds_ds)  # Convert `tf.data.Dataset` to Python generator

    for sample in tqdm(tfds_numpy):
        for col in sample:
            if col not in ds.tensors:
                ds.create_tensor(col)
            ds[col].extend(sample[col])
    return ds



import random
import time
import hashlib
import pickle
from typing import Any, Dict, Optional
import warnings
from hub.client.log import logger
from hub.constants import FIRST_COMMIT_ID
from hub.core.fast_forwarding import ffw_dataset_meta
from hub.core.meta.dataset_meta import DatasetMeta
from hub.core.storage.hub_memory_object import HubMemoryObject
from hub.core.version_control.commit_diff import CommitDiff
from hub.core.version_control.dataset_diff import DatasetDiff
from hub.core.version_control.commit_node import CommitNode  # type: ignore
from hub.core.version_control.commit_chunk_set import CommitChunkSet  # type: ignore
from hub.core.storage import LRUCache
from hub.core.lock import Lock
from hub.util.exceptions import CheckoutError, CommitError
from hub.util.keys import (
    get_chunk_id_encoder_key,
    get_creds_encoder_key,
    get_sequence_encoder_key,
    get_dataset_diff_key,
    get_dataset_info_key,
    get_dataset_meta_key,
    get_tensor_commit_chunk_set_key,
    get_tensor_commit_diff_key,
    get_tensor_info_key,
    get_tensor_meta_key,
    get_tensor_tile_encoder_key,
    get_version_control_info_key,
    get_version_control_info_key_old,
    get_version_control_info_lock_key,
)
from hub.util.remove_cache import get_base_storage
from hub.hooks import dataset_committed
from datetime import datetime
import json


def _version_info_to_json(info):
    commit_node_map, branch_commit_map = (
        info["commit_node_map"],
        info["branch_commit_map"],
    )
    commits = {}
    for commit, node in commit_node_map.items():
        commits[commit] = {
            "branch": node.branch,
            "parent": node.parent.commit_id if node.parent else None,
            "children": [c.commit_id for c in node.children],
            "commit_message": node.commit_message,
            "commit_time": node.commit_time.timestamp() if node.commit_time else None,
            "commit_user_name": node.commit_user_name,
        }
    return {
        "commits": commits,
        "branches": branch_commit_map,
    }


def _version_info_from_json(info):
    commits, branch_commit_map = info["commits"], info["branches"]
    commit_node_map = {}
    stack = [FIRST_COMMIT_ID]
    while stack:
        commit_id = stack.pop()
        commit_data = commits[commit_id]
        node = CommitNode(commit_data["branch"], commit_id)
        node.commit_message = commit_data["commit_message"]
        commit_time = commit_data["commit_time"]
        node.commit_time = (
            None if commit_time is None else datetime.fromtimestamp(commit_time)
        )
        node.commit_user_name = commit_data["commit_user_name"]
        parent = commit_data["parent"]
        if parent:
            commit_node_map[parent].add_child(node)
        commit_node_map[commit_id] = node
        stack += commit_data["children"]
    return {
        "commit_node_map": commit_node_map,
        "branch_commit_map": branch_commit_map,
    }


def generate_hash() -> str:
    hsh = hashlib.sha1()
    hsh.update(str(time.time()).encode("utf-8"))
    hsh.update(random.randrange(0, 1000000).to_bytes(4, "big"))
    return hsh.hexdigest()


def commit(dataset, message: str = None, hash: Optional[str] = None) -> None:
    """Modifies the version state to reflect the commit and also copies required data to the new commit directory."""
    storage = dataset.storage
    version_state = dataset.version_state
    storage.check_readonly()
    # if not the head node, checkout to an auto branch that is newly created
    auto_checkout(dataset)
    stored_commit_node: CommitNode = version_state["commit_node"]
    stored_commit_id = version_state["commit_id"]
    if hash:
        if hash in version_state["commit_node_map"]:
            raise CommitError(f"Commit {hash} already exists")
        version_state["commit_id"] = hash
    else:
        version_state["commit_id"] = generate_hash()
    new_node = CommitNode(version_state["branch"], version_state["commit_id"])
    version_state["commit_node"].add_successor(new_node, message)
    version_state["commit_node"] = new_node
    version_state["branch_commit_map"][version_state["branch"]] = version_state[
        "commit_id"
    ]
    version_state["commit_node_map"][version_state["commit_id"]] = new_node
    save_version_info(version_state, storage)
    copy_metas(stored_commit_id, version_state["commit_id"], storage, version_state)
    create_commit_chunk_sets(version_state["commit_id"], storage, version_state)
    discard_old_metas(stored_commit_id, storage, version_state["full_tensors"])
    load_meta(dataset)

    commit_time = stored_commit_node.commit_time
    commit_message = stored_commit_node.commit_message
    author = stored_commit_node.commit_user_name
    dataset._send_commit_event(
        commit_message=commit_message, commit_time=commit_time, author=author
    )
    dataset_committed(dataset)


def checkout(
    dataset,
    address: str,
    create: bool = False,
    hash: Optional[str] = None,
) -> None:
    """Modifies the version state to reflect the checkout and also copies required data to the new branch directory if a new one is being created."""
    storage = dataset.storage
    version_state = dataset.version_state
    original_commit_id = version_state["commit_id"]

    if address in version_state["branch_commit_map"].keys():
        if create:
            raise CheckoutError(f"Can't create new branch, '{address}' already exists.")
        new_commit_id = version_state["branch_commit_map"][address]
        if original_commit_id == new_commit_id:
            return
        if not storage.read_only:
            storage.flush()
        version_state["commit_id"] = new_commit_id
        version_state["commit_node"] = version_state["commit_node_map"][new_commit_id]
        version_state["branch"] = address

    elif address in version_state["commit_node_map"].keys():
        if create:
            raise CheckoutError(
                f"Can't create new branch, commit '{address}' already exists."
            )
        if address == original_commit_id:
            return
        if not storage.read_only:
            storage.flush()
        version_state["commit_id"] = address
        version_state["commit_node"] = version_state["commit_node_map"][address]
        version_state["branch"] = version_state["commit_node"].branch
    elif create:
        storage.check_readonly()
        # if the original commit is head of the branch, auto commit and checkout to original commit before creating new branch
        auto_commit(dataset, f"auto commit before checkout to {address}")
        if hash:
            if hash in version_state["commit_node_map"]:
                raise CommitError(f"Commit {hash} already exists")
            new_commit_id = hash
        else:
            new_commit_id = generate_hash()
        new_node = CommitNode(address, new_commit_id)
        version_state["commit_node"].add_child(new_node)
        version_state["commit_id"] = new_commit_id
        version_state["commit_node"] = new_node
        version_state["branch"] = address
        version_state["commit_node_map"][new_commit_id] = new_node
        version_state["branch_commit_map"][address] = new_commit_id
        save_version_info(version_state, storage)
        copy_metas(original_commit_id, new_commit_id, storage, version_state)
        create_commit_chunk_sets(new_commit_id, storage, version_state)
        dataset._send_branch_creation_event(address)
    else:
        raise CheckoutError(
            f"Address {address} not found. If you want to create a new branch, use checkout with create=True"
        )

    discard_old_metas(
        original_commit_id,
        storage,
        version_state["full_tensors"],
    )
    load_meta(dataset)


def copy_metas(
    src_commit_id: str,
    dest_commit_id: str,
    storage: LRUCache,
    version_state: Dict[str, Any],
) -> None:
    """Copies meta data from one commit to another."""
    initial_autoflush = storage.autoflush
    storage.autoflush = False

    tensors = version_state["full_tensors"]
    src_dataset_meta_key = get_dataset_meta_key(src_commit_id)
    dest_dataset_meta_key = get_dataset_meta_key(dest_commit_id)
    src_dataset_meta = storage[src_dataset_meta_key]
    dest_dataset_meta = convert_to_bytes(src_dataset_meta)
    storage[dest_dataset_meta_key] = dest_dataset_meta

    try:
        src_dataset_info_key = get_dataset_info_key(src_commit_id)
        dest_dataset_info_key = get_dataset_info_key(dest_commit_id)
        src_dataset_info = storage[src_dataset_info_key]
        dest_dataset_info = convert_to_bytes(src_dataset_info)
        storage[dest_dataset_info_key] = dest_dataset_info
    except KeyError:
        pass

    tensor_list = list(tensors.keys())

    for tensor in tensor_list:
        src_tensor_meta_key = get_tensor_meta_key(tensor, src_commit_id)
        dest_tensor_meta_key = get_tensor_meta_key(tensor, dest_commit_id)
        src_tensor_meta = storage[src_tensor_meta_key]
        dest_tensor_meta = convert_to_bytes(src_tensor_meta)
        storage[dest_tensor_meta_key] = dest_tensor_meta

        try:
            src_chunk_id_encoder_key = get_chunk_id_encoder_key(tensor, src_commit_id)
            dest_chunk_id_encoder_key = get_chunk_id_encoder_key(tensor, dest_commit_id)
            src_chunk_id_encoder = storage[src_chunk_id_encoder_key]
            dest_chunk_id_encoder = convert_to_bytes(src_chunk_id_encoder)
            storage[dest_chunk_id_encoder_key] = dest_chunk_id_encoder
        except KeyError:
            pass

        try:
            src_tile_encoder_key = get_tensor_tile_encoder_key(tensor, src_commit_id)
            dest_tile_encoder_key = get_tensor_tile_encoder_key(tensor, dest_commit_id)
            src_tile_encoder = storage[src_tile_encoder_key]
            dest_tile_encoder = convert_to_bytes(src_tile_encoder)
            storage[dest_tile_encoder_key] = dest_tile_encoder
        except KeyError:
            pass

        try:
            src_sequence_encoder_key = get_sequence_encoder_key(tensor, src_commit_id)
            dest_sequence_encoder_key = get_sequence_encoder_key(tensor, dest_commit_id)
            src_sequence_encoder = storage[src_sequence_encoder_key]
            dest_sequence_encoder = convert_to_bytes(src_sequence_encoder)
            storage[dest_sequence_encoder_key] = dest_sequence_encoder
        except KeyError:
            pass

        try:
            src_creds_encoder_key = get_creds_encoder_key(tensor, src_commit_id)
            dest_creds_encoder_key = get_creds_encoder_key(tensor, dest_commit_id)
            src_creds_encoder = storage[src_creds_encoder_key]
            dest_creds_encoder = convert_to_bytes(src_creds_encoder)
            storage[dest_creds_encoder_key] = dest_creds_encoder
        except KeyError:
            pass

        try:
            src_tensor_info_key = get_tensor_info_key(tensor, src_commit_id)
            dest_tensor_info_key = get_tensor_info_key(tensor, dest_commit_id)
            src_tensor_info = storage[src_tensor_info_key]
            dest_tensor_info = convert_to_bytes(src_tensor_info)
            storage[dest_tensor_info_key] = dest_tensor_info
        except KeyError:
            pass

    storage.autoflush = initial_autoflush
    storage.flush()


def create_commit_chunk_sets(
    dest_commit_id: str,
    storage: LRUCache,
    version_state: Dict[str, Any],
) -> None:
    """Creates commit chunk sets for all tensors in new commit."""
    tensor_list = version_state["full_tensors"].keys()
    for tensor in tensor_list:
        key = get_tensor_commit_chunk_set_key(tensor, dest_commit_id)
        storage[key] = CommitChunkSet()


def discard_old_metas(
    src_commit_id: str,
    storage: LRUCache,
    tensors: Dict,
):
    """Discards the metas of the previous commit from cache, during checkout or when a new commit is made."""
    all_src_keys = []
    src_dataset_meta_key = get_dataset_meta_key(src_commit_id)
    all_src_keys.append(src_dataset_meta_key)

    src_dataset_info_key = get_dataset_info_key(src_commit_id)
    all_src_keys.append(src_dataset_info_key)

    tensor_list = list(tensors.keys())

    for tensor in tensor_list:
        src_tensor_meta_key = get_tensor_meta_key(tensor, src_commit_id)
        all_src_keys.append(src_tensor_meta_key)

        src_chunk_id_encoder_key = get_chunk_id_encoder_key(tensor, src_commit_id)
        all_src_keys.append(src_chunk_id_encoder_key)

        src_tile_encoder_key = get_tensor_tile_encoder_key(tensor, src_commit_id)
        all_src_keys.append(src_tile_encoder_key)

        src_tensor_info_key = get_tensor_info_key(tensor, src_commit_id)
        all_src_keys.append(src_tensor_info_key)

    for key in all_src_keys:
        storage.dirty_keys.pop(key, None)
        if key in storage.lru_sizes:
            size = storage.lru_sizes.pop(key)
            storage.cache_used -= size
        try:
            del storage.cache_storage[key]
        except KeyError:
            pass


def _merge_commit_node_maps(map1, map2):
    merged_map = {}

    def _merge_node(commit_id):
        if commit_id in map1 and commit_id in map2:
            node1 = map1[commit_id]
            node2 = map2[commit_id]
            merged_node = CommitNode(node1.branch, node2.commit_id)

            for attr in ("commit_message", "commit_user_name", "commit_time"):
                setattr(merged_node, attr, getattr(node1, attr) or getattr(node2, attr))
            for child in set(
                [node.commit_id for node in node1.children]
                + [node.commit_id for node in node2.children]
            ):
                merged_node.add_child(_merge_node(child))
        else:
            if commit_id in map1:
                orig_node = map1[commit_id]
            else:
                orig_node = map2[commit_id]
            merged_node = orig_node.copy()
            for child in [node.commit_id for node in orig_node.children]:
                merged_node.add_child(_merge_node(child))
        merged_map[commit_id] = merged_node
        return merged_node

    _merge_node(FIRST_COMMIT_ID)
    return merged_map


def _merge_version_info(info1, info2):
    commit_node_map = _merge_commit_node_maps(
        info1["commit_node_map"], info2["commit_node_map"]
    )
    branch_commit_map = {}
    branch_commit_map.update(info1["branch_commit_map"])
    branch_commit_map.update(info2["branch_commit_map"])
    return {
        "commit_node_map": commit_node_map,
        "branch_commit_map": branch_commit_map,
    }


def save_version_info(version_state: Dict[str, Any], storage: LRUCache) -> None:
    """Saves the current version info to the storage."""
    storage = get_base_storage(storage)
    lock = Lock(storage, get_version_control_info_lock_key())
    lock.acquire(timeout=10, force=True)
    key = get_version_control_info_key()
    new_version_info = {
        "commit_node_map": version_state["commit_node_map"],
        "branch_commit_map": version_state["branch_commit_map"],
    }
    try:
        old_version_info = _version_info_from_json(
            json.loads(storage[key].decode("utf-8"))
        )
        version_info = _merge_version_info(old_version_info, new_version_info)
    except KeyError:
        try:
            old_version_info = pickle.loads(
                storage[get_version_control_info_key_old()]
            )  # backward compatiblity
            version_info = _merge_version_info(old_version_info, new_version_info)
        except KeyError:
            version_info = new_version_info
    storage[key] = json.dumps(_version_info_to_json(version_info)).encode("utf-8")
    lock.release()


def load_version_info(storage: LRUCache) -> Dict:
    try:
        return _version_info_from_json(
            json.loads(storage[get_version_control_info_key()].decode("utf-8"))
        )
    except KeyError:
        return pickle.loads(
            storage[get_version_control_info_key_old()]
        )  # backward compatiblity


def auto_checkout(dataset) -> bool:
    """Automatically checks out if current node is not the head node of the branch. This may happen either during commit/setitem/append/extend/create_tensor/delete_tensor/info updates."""
    version_state = dataset.version_state
    if not version_state["commit_node"].is_head_node:
        current_branch = version_state["branch"]
        auto_branch = f"auto_{generate_hash()}"
        logger.info(
            f"Automatically checking out to branch '{auto_branch}' as not currently at the head node of branch '{current_branch}'."
        )
        checkout(dataset, auto_branch, True)
        return True
    return False


def auto_commit(dataset, message: str) -> None:
    """Automatically commits to the current branch before a checkout to a newly created branch if the current node is the head node and has changes."""
    version_state = dataset.version_state
    commit_node = version_state["commit_node"]
    head = commit_node.is_head_node
    if not head:
        return

    if not current_commit_has_change(version_state, dataset.storage):
        parent_id = commit_node.parent.commit_id  # type: ignore
        checkout(dataset, parent_id, False)
        return

    original_commit_id = version_state["commit_id"]
    branch = version_state["branch"]
    logger.info(
        f"Auto commiting to branch '{branch}' before checkout as currently at head node."
    )
    commit(dataset, message)
    checkout(dataset, original_commit_id, False)


def current_commit_has_change(version_state: Dict[str, Any], storage: LRUCache) -> bool:
    return (
        current_commit_has_data(version_state, storage)
        or current_commit_has_info_modified(version_state, storage)
        or version_state["commit_id"] == FIRST_COMMIT_ID
    )


def current_commit_has_data(version_state: Dict[str, Any], storage: LRUCache) -> bool:
    """Checks if the current commit has any data present in it or not."""
    commit_id = version_state["commit_id"]
    try:
        dataset_diff_key = get_dataset_diff_key(commit_id)
        dataset_diff = storage.get_hub_object(dataset_diff_key, DatasetDiff)
        if dataset_diff.deleted or dataset_diff.renamed:
            return True
    except KeyError:
        pass

    for tensor in version_state["full_tensors"].keys():
        if commit_id == FIRST_COMMIT_ID:
            # if the first commit has even a single tensor i.e. it entered the for loop, it has data
            return True
        if commit_diff_exists(version_state, storage, tensor):
            # commit diff is created during tensor creation and append/extend/update
            return True
    return False


def current_commit_has_info_modified(
    version_state: Dict[str, Any], storage: LRUCache
) -> bool:
    commit_id = version_state["commit_id"]
    try:
        dataset_diff_key = get_dataset_diff_key(commit_id)
        dataset_diff = storage.get_hub_object(dataset_diff_key, DatasetDiff)
        if dataset_diff.info_updated:
            return True
    except KeyError:
        pass

    for tensor in version_state["full_tensors"].keys():
        try:
            tensor_diff_key = get_tensor_commit_diff_key(tensor, commit_id)
            tensor_diff = storage.get_hub_object(tensor_diff_key, CommitDiff)
            if tensor_diff.info_updated:
                return True
        except KeyError:
            pass

    return False


def commit_diff_exists(
    version_state: Dict[str, Any], storage: LRUCache, tensor: str
) -> bool:
    """Checks if the commit chunk set exists for the given tensor in the current commit."""
    try:
        commit_id = version_state["commit_id"]
        key = get_tensor_commit_diff_key(tensor, commit_id)
        storage[key]
        return True
    except KeyError:
        return False


def load_meta(dataset):
    """Loads the meta info for the version state."""
    from hub.core.tensor import Tensor

    version_state = dataset.version_state
    storage: LRUCache = dataset.storage
    storage.clear_hub_objects()
    meta_key = get_dataset_meta_key(version_state["commit_id"])
    meta = storage.get_hub_object(meta_key, DatasetMeta)
    if not meta.tensor_names:  # backward compatibility
        meta.tensor_names = {key: key for key in meta.tensors}

    ffw_dataset_meta(meta)
    version_state["meta"] = meta

    storage.register_hub_object(meta_key, meta)
    _tensors = version_state["full_tensors"]
    _tensors.clear()
    _tensor_names = version_state["tensor_names"]
    _tensor_names.clear()
    _tensor_names.update(meta.tensor_names)

    for tensor_key in _tensor_names.values():
        _tensors[tensor_key] = Tensor(tensor_key, dataset)


def warn_node_checkout(commit_node: CommitNode, create: bool):
    """Throws a warning if there are no commits in a branch after checkout.
    This warning isn't thrown if the branch was newly created.
    """
    if not create and commit_node.is_head_node:
        branch = commit_node.branch
        parent = commit_node.parent
        if parent is None or parent.branch != branch:
            warnings.warn(
                f"The branch ({branch}) that you have checked out to, has no commits."
            )


def convert_to_bytes(inp):
    return inp.tobytes() if isinstance(inp, HubMemoryObject) else inp



"""
Hub's Weights and Biases integration allows you to track and improve reproducibility of your machine learning experiments.
Hub will automatically push all information required to reproduce the snapshot of the data like your dataset's URI, commit ID, and view IDs of any 
views that you have used in your training workflow.

Learn more about Weights and Biases `here <https://wandb.ai>`_.

Logging Dataset Creation
~~~~~~~~~~~~~~~~~~~~~~~~
If you create a Hub dataset using any of the functions mentioned in :ref:`creating-datasets`, just perform a commit on the dataset to log its 
creation on W&B.

>>> run = wandb.init(project="hub_wandb", job_type="dataset_upload")
>>> ds = hub.empty("hub://fayazrahman4u/my_dataset") # create dataset
>>> ds.create_tensor("images", htype="image", sample_compression="jpg") # create a tensor
>>> ds.images.append(hub.read("files/images/dog.jpg")) # add a sample
>>> ds.commit("creation") # commit -> trigger logging
>>> run.finish()

NOTE:
    If you created your dataset using :meth:`hub.deepcopy`, perform the commit only if you have head changes.

NOTE:
    If you make changes to an existing dataset, commit the changes with an active Weights and Biases run to log it's state.

Logging Dataset Read
~~~~~~~~~~~~~~~~~~~~
A dataset read will be logged if you iterate over a dataset or call :meth:`Dataset.pytorch() <hub.core.dataset.Dataset.pytorch>` 
or :meth:`Tensor.numpy() <hub.core.tensor.Tensor.numpy>` on its tensors.

>>> run = wandb.init(project="hub_wandb", job_type="torch dataloader")
>>> train_loader = ds.pytorch()
>>> run.finish()

>>> run = wandb.init(project="hub_wandb", job_type="iteration")
>>> for sample in ds:
>>>     print(sample["images"].shape)
>>> run.finish()
"""

from typing import Dict, Set
from hub.util.tag import process_hub_path
from hub.util.hash import hash_inputs
from hub.constants import WANDB_JSON_FILENMAE
from hub.hooks import (
    add_create_dataset_hook,
    add_load_dataset_hook,
    add_write_dataset_hook,
    add_read_dataset_hook,
    add_commit_dataset_hook,
)
import importlib
import sys
import os
import json
import warnings
import hub

_WANDB_INSTALLED = bool(importlib.util.find_spec("wandb"))


def wandb_run():
    if not hub.constants.WANDB_INTEGRATION_ENABLED:
        return
    return getattr(sys.modules.get("wandb"), "run", None)


_READ_DATASETS: Dict[str, Set[str]] = {}
_WRITTEN_DATASETS: Dict[str, Set[str]] = {}


def dataset_created(ds):
    pass


def dataset_loaded(ds):
    pass


def artifact_name_from_ds_path(ds) -> str:
    path = ds.path
    hash = hash_inputs(path)
    if path.startswith("hub://"):
        _, org, ds_name, _ = process_hub_path(path)
        artifact_name = f"hub-{org}-{ds_name}"
        if "/.queries/" in path:
            vid = path.split("/.queries/", 1)[1]
            artifact_name += f"-view-{vid}"
    else:
        pfix = path.split("://", 1)[0] if "://" in path else "local"
        artifact_name = f"{pfix}"
    artifact_name += f"-commit-{ds.commit_id}"
    artifact_name += f"-{hash[:8]}"
    return artifact_name


def artifact_from_ds(ds):
    import wandb

    path = ds.path
    name = artifact_name_from_ds_path(ds)
    artifact = wandb.Artifact(name, "dataset")
    if "://" not in path and os.path.exists(path):
        path = "file://" + path
    artifact.add_reference(path, name="url")
    return artifact


def _is_public(ds_path):
    return True
    # TODO: We need api for this.
    try:
        hub.load(
            ds_path,
            token=hub.client.client.HubBackendClient(token="").request_auth_token(
                username="public", password=""
            ),
        )
        return True
    except Exception:
        return False


def get_ds_key(ds):
    entry = getattr(ds, "_view_entry", None)
    if entry:
        return hash_inputs(entry)
    return (hash_inputs(ds.path, ds.commit_id),)


def dataset_config(ds):
    if hasattr(ds, "_view_entry"):
        entry = ds._view_entry
        source_ds_path = entry.source_dataset_path
        commit_id = entry.info["source-dataset-version"]
        vid = entry.id
        ret = {
            "Dataset": source_ds_path,
            "Commit ID": commit_id,
            "View ID": vid,
        }
        if source_ds_path.startswith("hub://") and ds.path.startswith("hub://"):
            ret["URL"] = _plat_url(ds)
        q = entry.query
        if q:
            ret["Query"] = q
        if entry.virtual:
            ret["Index"] = ds.index.to_json()
        else:
            ret["Index"] = list(ds.sample_indices)
        return ret

    ret = {
        "Dataset": ds.path,
        "Commit ID": ds.commit_id,
    }
    if ds.path.startswith("hub://"):
        ret["URL"] = _plat_url(ds)
    if not ds.index.is_trivial():
        ret["Index"] = ds.index.to_json()
    q = getattr(ds, "_query", None)
    if q:
        ret["Query"] = q
    return ret


def log_dataset(dsconfig):
    # TODO: This is disabled until the embedded visualizer is actually useful for users.
    return
    url = dsconfig.get("URL")
    if not url:
        return
    import wandb

    run = wandb.run
    url_prefix = "https://app.activeloop.ai/"
    url = url[len(url_prefix) :]
    # TODO : commit and view id are not supported by visualizer. Remove below line once they are supported.
    url = "/".join(url.split("/")[:2])
    run.log({f"Hub Dataset - {url}": wandb.Html(_viz_html("hub://" + url))}, step=0)


def dataset_written(ds):
    pass


def dataset_committed(ds):
    run = wandb_run()
    key = get_ds_key(ds)
    if run:
        if run.id not in _WRITTEN_DATASETS:
            _WRITTEN_DATASETS.clear()
            _WRITTEN_DATASETS[run.id] = {}
        keys = _WRITTEN_DATASETS[run.id]
        if key not in keys:
            keys[key] = None
            try:
                output_datasets = run.config.output_datasets
            except (KeyError, AttributeError):
                output_datasets = []
            dsconfig = dataset_config(ds)
            output_datasets.append(dsconfig)
            log_dataset(dsconfig)
            run.config.output_datasets = output_datasets
            artifact = artifact_from_ds(ds)
            wandb_info = read_json(ds)
            try:
                commits = wandb_info["commits"]
            except KeyError:
                commits = {}
                wandb_info["commits"] = commits
            info = {}
            commits[ds.commit_id] = info
            info["created-by"] = {
                "run": {
                    "entity": run.entity,
                    "project": run.project,
                    "id": run.id,
                    "url": run.url,
                },
                "artifact": artifact.name,
            }
            write_json(ds, wandb_info)
            run.log_artifact(artifact)


def _filter_input_datasets(input_datasets):
    ret = []
    for i, dsconfig in enumerate(input_datasets):
        if "Index" not in dsconfig:
            rm = False
            for j, dsconfig2 in enumerate(input_datasets):
                if (
                    i != j
                    and dsconfig2["Dataset"] == dsconfig["Dataset"]
                    and dsconfig2["Commit ID"] == dsconfig["Commit ID"]
                ):
                    rm = True
                    break
            if not rm:
                ret.append(dsconfig)
        else:
            ret.append(dsconfig)
    return ret


def dataset_read(ds):
    run = wandb_run()
    if not run:
        return
    if run.id not in _READ_DATASETS:
        _READ_DATASETS.clear()
        _READ_DATASETS[run.id] = {}
    keys = _READ_DATASETS[run.id]
    key = get_ds_key(ds)
    idx = ds.index.to_json()
    if key not in keys or idx not in keys[key]:
        if key not in keys:
            keys[key] = []
        keys[key].append(idx)
        try:
            input_datasets = run.config.input_datasets
        except (KeyError, AttributeError):
            input_datasets = []
        dsconfig = dataset_config(ds)
        if dsconfig not in input_datasets:
            input_datasets.append(dsconfig)
            input_datasets = _filter_input_datasets(input_datasets)
            for config in input_datasets:
                log_dataset(config)
            run.config.input_datasets = input_datasets
        if run._settings.mode != "online":
            return
        if hasattr(ds, "_view_entry"):
            ds = ds._view_entry._src_ds
        wandb_info = read_json(ds).get("commits", {}).get(ds.commit_id)
        if wandb_info:
            try:
                run_and_artifact = wandb_info["created-by"]
                run_info = run_and_artifact["run"]
                artifact = run_and_artifact["artifact"]
                artifact_path = (
                    f"{run_info['entity']}/{run_info['project']}/{artifact}:latest"
                )
                run.use_artifact(artifact_path)
            except Exception as e:
                warnings.warn(
                    f"Wandb integration: Error while using wandb artifact: {e}"
                )
        else:
            # For datasets that were not created during a wandb run,
            # we want to "use" an artifact that is not logged by any run.
            # This is not possible with wandb yet.

            # artifact = artifact_from_ds(ds)
            # run.use_artifact(artifact)
            pass


def _viz_html(hub_path: str):
    if _is_public(hub_path):
        return f"""<iframe width="100%" height="100%" sandbox="allow-same-origin allow-scripts allow-popups allow-forms" src="https://app.activeloop.ai/visualizer/iframe?url={hub_path}" />"""
    return f"""
      <div id='container'></div>
  <script src="https://app.activeloop.ai/visualizer/vis.js"></script>
  <script>
    let container = document.getElementById('container')

    window.vis.visualize('{hub_path}', null, null, container, {{
      requireSignin: true
    }})
  </script>
    """


def _plat_url(ds, http=True):
    prefix = "https://app.activeloop.ai/" if http else "hub://"
    if hasattr(ds, "_view_entry"):
        entry = ds._view_entry
        _, org, ds_name, _ = process_hub_path(entry.source_dataset_path)
        commit_id = entry.info["source-dataset-version"]
        return f"{prefix}{org}/{ds_name}/{commit_id}?view={entry.id}"
    _, org, ds_name, _ = process_hub_path(ds.path)
    ret = f"{prefix}{org}/{ds_name}"
    if ds.commit_id:
        ret += f"/{ds.commit_id}"
    return ret


def link_html(hub_path):
    return f"""<a href="{_plat_url(hub_path)}">{hub_path}</a>"""


if _WANDB_INSTALLED:
    add_create_dataset_hook(dataset_created, "wandb_dataset_create")
    add_load_dataset_hook(dataset_loaded, "wandb_dataset_load")
    add_write_dataset_hook(dataset_written, "wandb_dataset_write")
    add_read_dataset_hook(dataset_read, "wandb_dataset_read")
    add_commit_dataset_hook(dataset_committed, "wandb_dataset_commit")


def read_json(ds):
    try:
        return json.loads(ds.base_storage[WANDB_JSON_FILENMAE].decode("utf-8"))
    except KeyError:
        return {}


def write_json(ds, dat):
    ds.base_storage[WANDB_JSON_FILENMAE] = json.dumps(dat).encode("utf-8")



import os
import pathlib
import posixpath
import shutil
from typing import Optional, Set

from hub.core.storage.provider import StorageProvider
from hub.util.exceptions import (
    DirectoryAtPathException,
    FileAtPathException,
    PathNotEmptyException,
)


class LocalProvider(StorageProvider):
    """Provider class for using the local filesystem."""

    def __init__(self, root: str):
        """Initializes the LocalProvider.

        Example:

            >>> local_provider = LocalProvider("/home/ubuntu/Documents/")

        Args:
            root (str): The root of the provider. All read/write request keys will be appended to root."

        Raises:
            FileAtPathException: If the root is a file instead of a directory.
        """
        if os.path.isfile(root):
            raise FileAtPathException(root)
        self.root = root
        self.files: Optional[Set[str]] = None
        self._all_keys()

    def subdir(self, path: str):
        return self.__class__(os.path.join(self.root, path))

    def __getitem__(self, path: str):
        """Gets the object present at the path within the given byte range.

        Example:

            >>> local_provider = LocalProvider("/home/ubuntu/Documents/")
            >>> my_data = local_provider["abc.txt"]

        Args:
            path (str): The path relative to the root of the provider.

        Returns:
            bytes: The bytes of the object present at the path.

        Raises:
            KeyError: If an object is not found at the path.
            DirectoryAtPathException: If a directory is found at the path.
            Exception: Any other exception encountered while trying to fetch the object.
        """
        try:
            full_path = self._check_is_file(path)
            with open(full_path, "rb") as file:
                return file.read()
        except DirectoryAtPathException:
            raise
        except FileNotFoundError:
            raise KeyError(path)

    def __setitem__(self, path: str, value: bytes):
        """Sets the object present at the path with the value

        Example:

            >>> local_provider = LocalProvider("/home/ubuntu/Documents/")
            >>> local_provider["abc.txt"] = b"abcd"

        Args:
            path (str): the path relative to the root of the provider.
            value (bytes): the value to be assigned at the path.

        Raises:
            Exception: If unable to set item due to directory at path or permission or space issues.
            FileAtPathException: If the directory to the path is a file instead of a directory.
            ReadOnlyError: If the provider is in read-only mode.
        """
        self.check_readonly()
        full_path = self._check_is_file(path)
        directory = os.path.dirname(full_path)
        if os.path.isfile(directory):
            raise FileAtPathException(directory)
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)
        with open(full_path, "wb") as file:
            file.write(value)
        if self.files is not None:
            self.files.add(path)

    def __delitem__(self, path: str):
        """Delete the object present at the path.

        Example:

            >>> local_provider = LocalProvider("/home/ubuntu/Documents/")
            >>> del local_provider["abc.txt"]

        Args:
            path (str): the path to the object relative to the root of the provider.

        Raises:
            KeyError: If an object is not found at the path.
            DirectoryAtPathException: If a directory is found at the path.
            Exception: Any other exception encountered while trying to fetch the object.
            ReadOnlyError: If the provider is in read-only mode.
        """
        self.check_readonly()
        try:
            full_path = self._check_is_file(path)
            os.remove(full_path)
            if self.files is not None:
                self.files.discard(path)
        except DirectoryAtPathException:
            raise
        except FileNotFoundError:
            raise KeyError

    def __iter__(self):
        """Generator function that iterates over the keys of the provider.

        Example:

            >>> local_provider = LocalProvider("/home/ubuntu/Documents/")
            >>> for my_data in local_provider:
            ...    pass

        Yields:
            str: the path of the object that it is iterating over, relative to the root of the provider.
        """
        yield from self._all_keys()

    def __len__(self):
        """Returns the number of files present inside the root of the provider.

        Example:

            >>> local_provider = LocalProvider("/home/ubuntu/Documents/")
            >>> len(local_provider)

        Returns:
            int: the number of files present inside the root.
        """
        return len(self._all_keys())

    def _all_keys(self, refresh: bool = False) -> Set[str]:
        """Lists all the objects present at the root of the Provider.

        Args:
            refresh (bool): refresh keys

        Returns:
            set: set of all the objects found at the root of the Provider.
        """
        if self.files is None or refresh:
            full_path = os.path.expanduser(self.root)
            key_set = set()
            for root, dirs, files in os.walk(full_path):
                for file in files:
                    key_set.add(
                        posixpath.relpath(
                            posixpath.join(pathlib.Path(root).as_posix(), file),
                            pathlib.Path(full_path).as_posix(),
                        )
                    )
            self.files = key_set
        return self.files

    def _check_is_file(self, path: str):
        """Checks if the path is a file. Returns the full_path to file if True.

        Args:
            path (str): the path to the object relative to the root of the provider.

        Returns:
            str: the full path to the requested file.

        Raises:
            DirectoryAtPathException: If a directory is found at the path.
        """
        full_path = posixpath.join(self.root, path)
        full_path = os.path.expanduser(full_path)
        full_path = str(pathlib.Path(full_path))
        if os.path.isdir(full_path):
            raise DirectoryAtPathException
        return full_path

    def clear(self, prefix=""):
        """Deletes ALL data with keys having given prefix on the local machine (under self.root). Exercise caution!"""
        self.check_readonly()
        full_path = os.path.expanduser(self.root)
        if prefix and self.files:
            self.files = set(file for file in self.files if not file.startswith(prefix))
            full_path = os.path.join(full_path, prefix)
        else:
            self.files = set()
        if os.path.exists(full_path):
            shutil.rmtree(full_path)

    def rename(self, path):
        """Renames root folder"""
        if os.path.isfile(path) or (os.path.isdir(path) and len(os.listdir(path)) > 0):
            raise PathNotEmptyException(use_hub=False)
        os.rename(self.root, path)
        self.root = path

    def __contains__(self, key) -> bool:
        full_path = self._check_is_file(key)
        return os.path.exists(full_path)

    def __getstate__(self):
        return self.root

    def __setstate__(self, state):
        self.__init__(state)

    def get_presigned_url(self, key: str) -> str:
        return os.path.join(self.root, key)

    def get_object_size(self, key: str) -> int:
        return os.stat(os.path.join(self.root, key)).st_size

    def get_bytes(
        self,
        path: str,
        start_byte: Optional[int] = None,
        end_byte: Optional[int] = None,
    ):
        try:
            full_path = self._check_is_file(path)
            with open(full_path, "rb") as file:
                if start_byte is not None:
                    file.seek(start_byte)
                if end_byte is None:
                    return file.read()
                else:
                    return file.read(end_byte - (start_byte or 0))
        except DirectoryAtPathException:
            raise
        except FileNotFoundError:
            raise KeyError(path)



from typing import Tuple
import numpy as np
import gzip

from hub.core.tiling.serialize import break_into_tiles, serialize_tiles
from hub.core.tiling.deserialize import coalesce_tiles, np_list_to_sample


def _get_arange_sample(shape: Tuple[int, ...]) -> np.ndarray:
    return np.arange(np.prod(shape)).reshape(*shape)


def test_break_into_tiles():
    sample = _get_arange_sample((2, 5))
    tile_shape = (3, 3)

    tiles = break_into_tiles(sample, tile_shape)
    assert tiles.shape == (1, 2)
    np.testing.assert_array_equal(tiles[0, 0], np.array([[0, 1, 2], [5, 6, 7]]))
    np.testing.assert_array_equal(tiles[0, 1], np.array([[3, 4], [8, 9]]))

    coalesced_sample = coalesce_tiles(tiles, tile_shape, sample.shape, sample.dtype)
    np.testing.assert_array_equal(sample, coalesced_sample)


def test_serialize_tiles():
    sample = _get_arange_sample((2, 5))
    tile_shape = (3, 3)

    tiles = break_into_tiles(sample, tile_shape)
    shapes = [t.shape for _, t in np.ndenumerate(tiles)]
    serialized_tiles = serialize_tiles(tiles, lambda x: memoryview(x.tobytes()))
    assert serialized_tiles.shape == tiles.shape

    flattened_tiles = serialized_tiles.reshape((serialized_tiles.size,))
    tiled_arrays = [
        np.frombuffer(x, dtype=sample.dtype).reshape(sh)
        for x, sh in zip(flattened_tiles, shapes)
    ]
    coalesced_sample = np_list_to_sample(
        tiled_arrays, sample.shape, tile_shape, tiles.shape, sample.dtype
    )
    np.testing.assert_array_equal(sample, coalesced_sample)


def test_serialize_tiles_gzip():
    sample = _get_arange_sample((2, 5))
    tile_shape = (3, 3)

    tiles = break_into_tiles(sample, tile_shape)
    shapes = [t.shape for _, t in np.ndenumerate(tiles)]
    gzip_compress = lambda x: memoryview(gzip.compress(x.tobytes()))
    serialized_tiles = serialize_tiles(tiles, gzip_compress)
    assert serialized_tiles.shape == tiles.shape

    flattened_tiles = serialized_tiles.reshape((serialized_tiles.size,))
    gzip_decompress = lambda x: np.frombuffer(gzip.decompress(x), dtype=sample.dtype)
    tiled_arrays = [
        gzip_decompress(x).reshape(sh) for x, sh in zip(flattened_tiles, shapes)
    ]

    coalesced_sample = np_list_to_sample(
        tiled_arrays, sample.shape, tile_shape, tiles.shape, sample.dtype
    )
    np.testing.assert_array_equal(sample, coalesced_sample)



import hub
import pytest
from pydicom.data import get_testdata_file
from pydicom import dcmread
from hub.util.exceptions import UnsupportedCompressionError


def test_dicom_basic(memory_ds):
    ds = memory_ds
    path = get_testdata_file("MR_small.dcm")
    with ds:
        ds.create_tensor("x", htype="dicom")
        with pytest.raises(UnsupportedCompressionError):
            ds.create_tensor("y", htype="dicom", sample_compression="jpg")
        dcm = hub.read(path)
        assert dcm.dtype == "int16"
        assert dcm.shape == (64, 64, 1)
        ds.x.append(dcm)
        assert ds.x.data()["value"].shape == (1, 64, 64, 1)
        dcm = hub.read(path, verify=True)
        assert dcm.dtype == "int16"
        assert dcm.shape == (64, 64, 1)
        ds.x.append(dcm)
    assert ds.x.dtype == "int16"
    arr = ds.x.numpy()
    assert arr.dtype == "int16"
    assert arr.shape == (2, 64, 64, 1)
    for item in dcmread(path):
        if not isinstance(item.value, bytes):
            assert item.keyword in ds.x[0].sample_info


def test_dicom_mixed_dtype(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("x", htype="dicom")
        dcm = hub.read(get_testdata_file("MR_small.dcm"))
        assert dcm.dtype == "int16"
        ds.x.append(dcm)
        dcm = hub.read(get_testdata_file("ExplVR_BigEnd.dcm"))
        assert dcm.dtype == "uint8"
        ds.x.append(dcm)
    arr = ds.x[:, :10, :10, :1].numpy()
    assert arr.dtype == "int16"



import numpy as np
from hub.constants import KB


def _assert_num_chunks(tensor, expected_num_chunks):
    chunk_engine = tensor.chunk_engine
    actual_num_chunks = chunk_engine.chunk_id_encoder.num_chunks
    assert actual_num_chunks == expected_num_chunks


def _create_tensors(ds):
    images = ds.create_tensor(
        "images",
        htype="image",
        sample_compression=None,
        max_chunk_size=32 * KB,
        tiling_threshold=16 * KB,
    )
    labels = ds.create_tensor(
        "labels", htype="class_label", max_chunk_size=32 * KB, tiling_threshold=16 * KB
    )
    return images, labels


def _append_tensors(images, labels):
    for i in range(100):
        x = np.ones((28, 28), dtype=np.uint8) * i
        y = np.uint32(i)

        images.append(x)
        labels.append(y)


def _extend_tensors(images, labels):
    images.extend(np.ones((100, 28, 28), dtype=np.uint8))
    labels.extend(np.ones(100, dtype=np.uint32))


def _clear_tensors(images, labels):
    images.clear()
    labels.clear()


def test_append(memory_ds):
    ds = memory_ds
    images, labels = _create_tensors(ds)

    _append_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 5)

    _append_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 10)

    _append_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 15)

    assert len(ds) == 300


def test_extend(memory_ds):
    ds = memory_ds
    images, labels = _create_tensors(ds)

    _extend_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 5)

    _extend_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 10)

    _extend_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 15)

    assert len(ds) == 300


def test_extend_and_append(memory_ds):
    ds = memory_ds
    images, labels = _create_tensors(ds)

    _extend_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 5)

    _append_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 10)

    _extend_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 15)

    _append_tensors(images, labels)

    _assert_num_chunks(labels, 1)
    _assert_num_chunks(images, 20)

    assert len(ds) == 400


def test_clear(memory_ds):
    ds = memory_ds
    images, labels = _create_tensors(ds)

    _clear_tensors(images, labels)

    _assert_num_chunks(labels, 0)
    _assert_num_chunks(images, 0)



import requests
from hub.core.compression import (
    compress_array,
    decompress_array,
    verify_compressed_file,
    read_meta_from_compressed_file,
    get_compression,
    _open_video,
    _read_metadata_from_vstream,
    _read_audio_meta,
    _read_point_cloud_meta,
)
from hub.compression import (
    get_compression_type,
    AUDIO_COMPRESSION,
    IMAGE_COMPRESSION,
    VIDEO_COMPRESSION,
    POINT_CLOUD_COMPRESSION,
)
from hub.util.exceptions import UnableToReadFromUrlError
from hub.util.exif import getexif
from hub.core.storage.provider import StorageProvider
from hub.util.path import get_path_type, is_remote_path
import numpy as np
from typing import Optional, Tuple, Union, Dict

from PIL import Image  # type: ignore
from PIL.ExifTags import TAGS  # type: ignore
from io import BytesIO

from hub.core.storage.s3 import S3Provider
from hub.core.storage.google_drive import GDriveProvider

try:
    from hub.core.storage.gcs import GCSProvider
except ImportError:
    GCSProvider = None  # type: ignore

import warnings


class Sample:
    path: Optional[str]

    def __init__(
        self,
        path: Optional[str] = None,
        array: np.ndarray = None,
        buffer: Union[bytes, memoryview] = None,
        compression: str = None,
        verify: bool = False,
        shape: Tuple[int] = None,
        dtype: Optional[str] = None,
        creds: Optional[Dict] = None,
        storage: Optional[StorageProvider] = None,
    ):
        """Represents a single sample for a tensor. Provides all important meta information in one place.

        Note:
            If ``self.is_lazy`` is ``True``, this :class:`Sample` doesn't actually have any data loaded. To read this data, simply try to read it into a numpy array (`sample.array`)

        Args:
            path (str): Path to a sample stored on the local file system that represents a single sample. If ``path`` is provided, ``array`` should not be.
                Implicitly makes ``self.is_lazy == True``.
            array (np.ndarray): Array that represents a single sample. If ``array`` is provided, ``path`` should not be. Implicitly makes ``self.is_lazy == False``.
            buffer: (bytes): Byte buffer that represents a single sample. If compressed, ``compression`` argument should be provided.
            compression (str): Specify in case of byte buffer.
            verify (bool): If a path is provided, verifies the sample if ``True``.
            shape (Tuple[int]): Shape of the sample.
            dtype (optional, str): Data type of the sample.
            creds (optional, Dict): Credentials for s3, gcp and http urls.
            storage (optional, StorageProvider): Storage provider.

        Raises:
            ValueError: Cannot create a sample from both a ``path`` and ``array``.
        """
        if path is None and array is None and buffer is None:
            raise ValueError("Must pass one of `path`, `array` or `buffer`.")

        self._compressed_bytes = {}
        self._uncompressed_bytes = None

        self._array = None
        self._typestr = None
        self._shape = shape or None
        self._dtype = dtype or None

        self.path = None
        self.storage = storage
        self._buffer = None
        self._creds = creds or {}
        self._verify = verify

        if path is not None:
            self.path = path
            self._compression = compression
            if self._verify:
                if self._compression is None:
                    self._compression = get_compression(path=self.path)
                compressed_bytes = self._read_from_path()
                if self._compression is None:
                    self._compression = get_compression(header=compressed_bytes[:32])
                self._shape, self._typestr = verify_compressed_file(compressed_bytes, self._compression)  # type: ignore

        if array is not None:
            self._array = array
            self._shape = array.shape  # type: ignore
            self._typestr = array.__array_interface__["typestr"]
            self._dtype = np.dtype(self._typestr).name
            self._compression = None

        if buffer is not None:
            self._compression = compression
            self._buffer = buffer
            if compression is None:
                self._uncompressed_bytes = buffer
            else:
                self._compressed_bytes[compression] = buffer
                if self._verify:
                    self._shape, self._typestr = verify_compressed_file(buffer, self._compression)  # type: ignore

        self.htype = None

    @property
    def buffer(self):
        if self._buffer is None and self.path is not None:
            self._read_from_path()
        if self._buffer is not None:
            return self._buffer
        return self.compressed_bytes(self.compression)

    @property
    def is_text_like(self):
        return self.htype in {"text", "list", "json"}

    @property
    def dtype(self):
        if self._dtype is None:
            self._read_meta()
            self._dtype = np.dtype(self._typestr).name
        return self._dtype

    @property
    def shape(self):
        self._read_meta()
        return self._shape

    @property
    def compression(self):
        if self._compression is None and self.path:
            self._read_meta()
        return self._compression

    def _load_dicom(self):
        if self._array is not None:
            return
        try:
            from pydicom import dcmread
        except ImportError:
            raise ModuleNotFoundError(
                "Pydicom not found. Install using `pip install pydicom`"
            )
        if self.path and get_path_type(self.path) == "local":
            dcm = dcmread(self.path)
        else:
            dcm = dcmread(BytesIO(self.buffer))
        self._array = dcm.pixel_array
        self._shape = self._array.shape
        self._typestr = self._array.__array_interface__["typestr"]

    def _read_meta(self, f=None):
        if self._shape is not None:
            return
        store = False
        if self._compression is None and self.path:
            self._compression = get_compression(path=self.path)
        if f is None:
            if self.path:
                if is_remote_path(self.path):
                    f = self._read_from_path()
                    self._buffer = f
                    store = True
                else:
                    f = self.path
            else:
                f = self._buffer
        self._compression, self._shape, self._typestr = read_meta_from_compressed_file(
            f, compression=self._compression
        )
        if store:
            self._compressed_bytes[self._compression] = f

    def _get_dicom_meta(self) -> dict:
        try:
            from pydicom import dcmread
            from pydicom.dataelem import RawDataElement
        except ImportError:
            raise ModuleNotFoundError(
                "Pydicom not found. Install using `pip install pydicom`"
            )
        if self.path and get_path_type(self.path) == "local":
            dcm = dcmread(self.path)
        else:
            dcm = dcmread(BytesIO(self.buffer))

        meta = {
            x.keyword: {
                "name": x.name,
                "tag": str(x.tag),
                "value": x.value
                if isinstance(x.value, (str, int, float))
                else x.to_json_dict(None, None).get("Value", ""),  # type: ignore
                "vr": x.VR,
            }
            for x in dcm
            if not isinstance(x.value, bytes)
        }
        return meta

    def _get_video_meta(self) -> dict:
        if self.path and get_path_type(self.path) == "local":
            container, vstream = _open_video(self.path)
        else:
            container, vstream = _open_video(self.buffer)
        _, duration, fps, timebase = _read_metadata_from_vstream(container, vstream)
        return {"duration": duration, "fps": fps, "timebase": timebase}

    def _get_audio_meta(self) -> dict:
        if self.path and get_path_type(self.path) == "local":
            info = _read_audio_meta(self.path)
        else:
            info = _read_audio_meta(self.buffer)
        return info

    def _get_point_cloud_meta(self) -> dict:
        if self.path and get_path_type(self.path) == "local":
            info = _read_point_cloud_meta(self.path)
        else:
            info = _read_point_cloud_meta(self.buffer)
        return info

    @property
    def is_lazy(self) -> bool:
        return self._array is None

    @property
    def is_empty(self) -> bool:
        return 0 in self.shape

    def _recompress(self, buffer: bytes, compression: str) -> bytes:
        if get_compression_type(self._compression) != IMAGE_COMPRESSION:
            raise ValueError(
                "Recompression with different format is only supported for images."
            )
        img = Image.open(BytesIO(buffer))
        if img.mode == "1":
            self._uncompressed_bytes = img.tobytes("raw", "L")
        else:
            self._uncompressed_bytes = img.tobytes()
        return compress_array(self.array, compression)

    def compressed_bytes(self, compression: Optional[str]) -> bytes:
        """Returns this sample as compressed bytes.

        Note:
            If this sample is pointing to a path and the requested ``compression`` is the same as it's stored in, the data is returned without re-compressing.

        Args:
            compression (Optional[str]): ``self.array`` will be compressed into this format. If ``compression`` is ``None``, return :meth:`uncompressed_bytes`.

        Returns:
            bytes: Bytes for the compressed sample. Contains all metadata required to decompress within these bytes.

        Raises:
            ValueError: On recompression of unsupported formats.
        """

        if compression is None:
            return self.uncompressed_bytes()  # type: ignore

        compressed_bytes = self._compressed_bytes.get(compression)
        if compressed_bytes is None:
            if self.path is not None:
                if self._compression is None:
                    self._compression = get_compression(path=self.path)
                compressed_bytes = self._read_from_path()
                if self._compression is None:
                    self._compression = get_compression(header=compressed_bytes[:32])
                if self._compression == compression:
                    if self._shape is None:
                        _, self._shape, self._typestr = read_meta_from_compressed_file(
                            compressed_bytes, compression=self._compression
                        )
                else:
                    compressed_bytes = self._recompress(compressed_bytes, compression)
            elif self._buffer is not None:
                compressed_bytes = self._recompress(self._buffer, compression)
            else:
                compressed_bytes = compress_array(self.array, compression)
            self._compressed_bytes[compression] = compressed_bytes
        return compressed_bytes

    def _decompress(self):
        if self._array is not None:
            if self._uncompressed_bytes is None:
                self._uncompressed_bytes = self._array.tobytes()
            return
        compression = self.compression
        if compression is None and self._buffer is not None:
            if self.is_text_like:
                from hub.core.serialize import bytes_to_text

                buffer = bytes(self._buffer)
                self._array = bytes_to_text(buffer, self.htype)
            else:
                self._array = np.frombuffer(self._buffer, dtype=self.dtype).reshape(
                    self.shape
                )

        else:
            if self.path and get_path_type(self.path) == "local":
                compressed = self.path
            else:
                compressed = self.buffer

            self._array = decompress_array(
                compressed, compression=compression, shape=self.shape, dtype=self.dtype
            )
            self._uncompressed_bytes = self._array.tobytes()
            self._typestr = self._array.__array_interface__["typestr"]
            self._dtype = np.dtype(self._typestr).name

    def uncompressed_bytes(self) -> Optional[bytes]:
        """Returns uncompressed bytes."""
        self._decompress()
        return self._uncompressed_bytes

    @property
    def array(self) -> np.ndarray:  # type: ignore
        """Return numpy array corresponding to the sample. Decompresses the sample if necessary.

        Example:

            >>> sample = hub.read("./images/dog.jpg")
            >>> arr = sample.array
            >>> arr.shape
            (323, 480, 3)
        """
        arr = self._array
        if arr is not None:
            return arr
        self._decompress()
        return self._array  # type: ignore

    def __str__(self):
        if self.is_lazy:
            return f"Sample(is_lazy=True, path={self.path})"

        return f"Sample(is_lazy=False, shape={self.shape}, compression='{self.compression}', dtype='{self.dtype}' path={self.path})"

    def __repr__(self):
        return str(self)

    def __array__(self, dtype=None):
        arr = self.array
        if dtype is not None:
            arr = arr.astype(dtype)
        return arr

    def __eq__(self, other):
        if self.path is not None and other.path is not None:
            return self.path == other.path
        return self.buffer == other.buffer

    def _read_from_path(self) -> bytes:  # type: ignore
        if self._buffer is None:
            path_type = get_path_type(self.path)
            if path_type == "local":
                self._buffer = self._read_from_local()
            elif path_type == "gcs":
                self._buffer = self._read_from_gcs()
            elif path_type == "s3":
                self._buffer = self._read_from_s3()
            elif path_type == "gdrive":
                self._buffer = self._read_from_gdrive()
            elif path_type == "http":
                self._buffer = self._read_from_http()
        return self._buffer  # type: ignore

    def _read_from_local(self) -> bytes:
        with open(self.path, "rb") as f:  # type: ignore
            return f.read()

    def _get_root_and_key(self, path):
        split_path = path.split("/", 2)
        if len(split_path) > 2:
            root, key = "/".join(split_path[:2]), split_path[2]
        else:
            root, key = split_path
        return root, key

    def _read_from_s3(self) -> bytes:
        assert self.path is not None
        if self.storage is not None:
            assert isinstance(self.storage, S3Provider)
            return self.storage.get_object_from_full_url(self.path)
        path = self.path.replace("s3://", "")  # type: ignore
        root, key = self._get_root_and_key(path)
        s3 = S3Provider(root, **self._creds)
        return s3[key]

    def _read_from_gcs(self) -> bytes:
        assert self.path is not None
        if GCSProvider is None:
            raise Exception(
                "GCP dependencies not installed. Install them with pip install hub[gcs]"
            )
        if self.storage is not None:
            assert isinstance(self.storage, GCSProvider)
            return self.storage.get_object_from_full_url(self.path)
        path = self.path.replace("gcp://", "").replace("gcs://", "")  # type: ignore
        root, key = self._get_root_and_key(path)
        gcs = GCSProvider(root, token=self._creds)
        return gcs[key]

    def _read_from_gdrive(self) -> bytes:
        assert self.path is not None
        gdrive = GDriveProvider("gdrive://", token=self._creds, makemap=False)
        return gdrive.get_object_from_full_url(self.path)

    def _read_from_http(self) -> bytes:
        assert self.path is not None
        if "Authorization" in self._creds:
            headers = {"Authorization": self._creds["Authorization"]}
        else:
            headers = {}
        result = requests.get(self.path, headers=headers)
        if result.status_code != 200:
            raise UnableToReadFromUrlError(self.path, result.status_code)
        return result.content

    def _getexif(self) -> dict:
        if self.path and get_path_type(self.path) == "local":
            img = Image.open(self.path)
        else:
            img = Image.open(BytesIO(self.buffer))
        try:
            return getexif(img)
        except Exception as e:
            warnings.warn(
                f"Error while reading exif data, possibly due to corrupt exif: {e}"
            )
            return {}

    @property
    def meta(self) -> dict:
        meta: Dict[str, Union[Dict, str]] = {}
        compression = self.compression
        compression_type = get_compression_type(compression)
        if compression == "dcm":
            meta.update(self._get_dicom_meta())
        elif compression_type == IMAGE_COMPRESSION:
            meta["exif"] = self._getexif()
        elif compression_type == VIDEO_COMPRESSION:
            meta.update(self._get_video_meta())
        elif compression_type == AUDIO_COMPRESSION:
            meta.update(self._get_audio_meta())
        elif compression_type == POINT_CLOUD_COMPRESSION:
            meta.update(self._get_point_cloud_meta())
        meta["shape"] = self.shape
        meta["format"] = self.compression
        if self.path:
            meta["filename"] = str(self.path)
        return meta


SampleValue = Union[np.ndarray, int, float, bool, Sample]



from typing import Dict, Tuple


class PartialReader:
    def __init__(self, cache, path: str, header_offset: int):
        self.cache = cache
        self.path = path
        self.data_fetched: Dict[Tuple[int, int], memoryview] = {}
        self.header_offset = header_offset

    def __getitem__(self, slice_: slice) -> memoryview:
        start = slice_.start + self.header_offset
        stop = slice_.stop + self.header_offset
        step = slice_.step
        assert start is not None and stop is not None
        assert step is None or step == 1
        slice_tuple = (start, stop)
        if slice_tuple not in self.data_fetched:
            self.data_fetched[slice_tuple] = memoryview(
                self.cache.get_bytes(self.path, start, stop)
            )
        return self.data_fetched[slice_tuple]

    def get_all_bytes(self) -> bytes:
        return self.cache.next_storage[self.path]
